/**
 * mod_trap_ifc.c
 * Copyright (C) 2013, 2014 Roman Vrana
 * Module for servicing TRAP intterface used for communication with filterd
 * program.
 */

#include "../ccci_msg.h"
#include "../cccid_mod.h"
#include "../cccid_conf.h"
#include "rid_mgmt.h"

#include <libtrap/trap.h>
#include <stdlib.h>

#define MAX_RID 65536
#define RULE_ADD 0
#define RULE_DEL 1

static
trap_module_info_t module_info = {
    "CCCI daemon for LI system -- TRAP interface\n",
    "Module for communication with SLIS system via CCCI message format.\n"
    "Module also passes commands for following filterd daemon.\n",
    0, // no TRAP interface for input
    1 // 1 TRAP output interface (CCCI message)
};

static
int trapifc_match(const union ccci_msg *req)
{
    // message will be processed in every situation --> return 0 every time
	return 0;
}

static
int trapifc_service(const union ccci_msg *req, union ccci_msg *res)
{

    memcpy(res, req, sizeof(union ccci_msg));

    // send to filterd
    trap_send(0, res, sizeof(union ccci_msg));


    // create response
    ccci_msg_setstatus(res, 1);

	return 1;
}

static
void trapifc_fini()
{
    trap_terminate();
    trap_finalize();
}

static
struct cccid_mod mod_trapifc = {
	.name    = "trap_ifc",
	.version = 0x0001,
	.match   = trapifc_match,
	.service = trapifc_service,
    .fini = trapifc_fini
};

static
void trapifc_init(struct cccid_conf *conf)
{
	if(cccid_mod_install(&mod_trapifc)) {
		syslog(LOG_ALERT, "Failed to install mod_%s 0x%" PRIx16,
				mod_trapifc.name, mod_trapifc.version);
	} else {
        const char* trap_string = cccid_conf_gets(conf, "mod.trap.ifcspec", NULL);

        if (trap_string == NULL) {
            syslog(LOG_ALERT, "TRAP module failure: No paramaters given.");
            return;
        }
        
        trap_ifc_spec_t ifc_spec;
        
        // check interface count (should be only one)
        int i = 0;
        while(trap_string[i] != ';') {
            i++;
        }

        if (i != 1) {
            syslog(LOG_ALERT, "TRAP module failure: Only 1 interface allowed");
        } else {
            ifc_spec.types = (char *) malloc(2 * sizeof(char));
            ifc_spec.types[0] = trap_string[0]; // type of interface (should be only one)
            ifc_spec.types[1] = '\0'; // type of interface (should be only one)

            // prepare interface parameters structure
            ifc_spec.params = (char **) malloc(sizeof(char *));

            // can use (trap_string + 2) since only characters before are type and ';'
            ifc_spec.params[0] = (char *) malloc (sizeof(char) * (strlen(trap_string + 2)));

            // copy to structure
            strcpy(ifc_spec.params[0], (trap_string + 2));

            // initialize trap interface
            int ret = trap_init(&module_info, ifc_spec);
            
            if (ret != TRAP_E_OK) {
                syslog(LOG_ALERT, "TRAP module failure: Unable to initialize.");
                return;
            }

            // set interface timeout to HALFWAIT mode
            // (wait if client connected, skip otherwise)
            trap_ifcctl(TRAPIFC_OUTPUT, 0, TRAPCTL_SETTIMEOUT, TRAP_WAIT);
            trap_ifcctl(TRAPIFC_OUTPUT, 0, TRAPCTL_BUFFERSWITCH, 0);

            // free the structure of ifc specification
            free(ifc_spec.types);
            free(ifc_spec.params[0]);
            free(ifc_spec.params);

            // ready to work
        }
    }
}

cccid_mod_init(trap_ifc, trapifc_init)
