/**
 * \file filter_control.c
 * \brief Functions and procedures for filter table maintenace
 * \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
 * \date 2014
 */
#include <stdlib.h>
#include <string.h>
#include <endian.h>

#include "filter_control.h"
#include "inet_offsets.h"
#include "ccci_struct/ccci_rule.h"

/**
 * \brief Function for filling flow key from incomming CCCI message.
 * 
 * This function takes data from CCCI message received from SLIS system
 * and converts it to lookup format 
 *
 * \param rule
 * \param key
 */
void filter_key_ccci(union ccci_rule *rule, sdm_flow_identifier_t *key)
{
    // create the key from CCCI
    key->elems.protocol = ccci_rule_getproto4(rule);
    key->elems.ip_version = ccci_rule_getipver(rule);

    uint16_t p1, p2;
    sdm_ip_t ip1, ip2;
    uint32_t ip4s, ip4d;

    key->elems.ip_version = ccci_rule_getipver(rule);
    if (ccci_rule_getipver(rule) == 4) {
        ip4s = ccci_rule_getip0(rule)->raw32[0];
        ip4d = ccci_rule_getip1(rule)->raw32[0];

        if (ip4s <= ip4d) {
            key->elems.src_ip._v4.d32[0] = ip4s;
            key->elems.dst_ip._v4.d32[0] = ip4d;
            key->elems.src_port = ccci_rule_getport0(rule);
            key->elems.dst_port = ccci_rule_getport1(rule);
        } else {
            key->elems.src_ip._v4.d32[0] = ip4d;
            key->elems.dst_ip._v4.d32[0] = ip4s;
            key->elems.src_port = ccci_rule_getport1(rule);
            key->elems.dst_port = ccci_rule_getport0(rule);
        }

#ifdef IPV6
    } else {
        if (memcmp(ccci_rule_getip0(rule), ccci_rule_getip1(rule), 16) > 0) {
            // Source is greater than destination -- swap IP and port
            key->elems.src_ip.v6.d64[0] = ccci_rule_getip1(rule)->raw64[0];
            key->elems.src_ip.v6.d64[1] = ccci_rule_getip1(rule)->raw64[1];
            key->elems.dst_ip.v6.d64[0] = ccci_rule_getip0(rule)->raw64[0];
            key->elems.dst_ip.v6.d64[1] = ccci_rule_getip0(rule)->raw64[1];

            key->elems.src_port = ccci_rule_getport1(rule);
            key->elems.dst_port = ccci_rule_getport0(rule);
        } else {
            // Keep order as it is
            key->elems.src_ip.v6.d64[0] = ccci_rule_getip0(rule)->raw64[0];
            key->elems.src_ip.v6.d64[1] = ccci_rule_getip0(rule)->raw64[1];
            key->elems.dst_ip.v6.d64[0] = ccci_rule_getip1(rule)->raw64[0];
            key->elems.dst_ip.v6.d64[1] = ccci_rule_getip1(rule)->raw64[1];

            key->elems.src_port = ccci_rule_getport0(rule);
            key->elems.dst_port = ccci_rule_getport1(rule);
        }
    }
#else
    }
#endif
#ifdef DEBUG
    printf("Inserted flow key:");
    for (int i = 0; i < 40; i++)
        printf("%2x ", key->raw[i]);
    printf("\b\n");
#endif
}

/**
* \brief Function for initializing filter sets.
*
* \param filter Filter set to be initialized.
*
* \return 0 on success 1 on flow filter init failure.
*/
int filter_init(filter_table_t *filter)
{
    int i = 0;

    pref_init(&(filter->v4_pref));
    pref_init(&(filter->v6_pref));
    filter->flow_filter = fhf_init(FLOW_FILTER_SIZE,
                                   sizeof(sdm_flow_identifier_t),
                                   sizeof(uint32_t));
#ifndef NO_PORT
    filter->port_filter = fhf_init(PORT_FILTER_SIZE,
                                   sizeof(sdm_ip_t),
                                   sizeof(uint32_t) * (UINT16_MAX + 1));
#endif

    if (filter->flow_filter == NULL || filter->port_filter == NULL)
        return 1;
    else
        return 0;
}

/**
 * \brief Procedure for adding prefix rule to filter table.
 *
 * Procedure takes prepared prefix filtering rule and puts it into 
 * corresponding prefix filter.
 *
 * \param filter Filter to use edit.
 * \param data Prefix rule to store.
 * \param ip_ver Version of ip address in the rule.
 */
void filter_add_prefix(filter_table_t *filter,
                       pref_item_t *data,
                       uint8_t ip_ver)
{
    if (ip_ver == 4) {
        pref_add(&(filter->v4_pref), data);
    } else {
        pref_add(&(filter->v6_pref), data);
    }
}

/**
 * \brief Procedure for adding flow filtering rule to filter table.
 *
 * Procedure takes prepared key of the flow and stores it in the hash
 * filter. If the item cannot be stored in hash table then the backup 
 * stash is used for storing the item.
 * 
 * \param filter Filter to edit.
 * \param data Data to store.
 *
 * \return 0 if everything goes well, -1 otherwise
 */
int filter_add_flow(filter_table_t *filter, filter_item_t *data)
{
    void *updated = NULL;
    int ins_ret = FHF_INSERT_FAILED;
    int8_t *data_lock = NULL;

    ins_ret = fhf_insert_own_or_update(filter->flow_filter, 
                                       data->key.raw,
                                       &data_lock,
                                       &updated);
    if (ins_ret == FHF_INSERT_FAILED) {
        // Item can be inserted or is already in filter
        *((uint32_t *) updated) = data->rid;
    }

    if (ins_ret != FHF_INSERT_FULL) {
        fhf_unlock_data(data_lock);
    }   

    return ins_ret;
}

#ifndef NO_PORT
/**
 * \brief Procedure for adding IP-port filtering rule to filter table.
 *
 * 
 * \param filter Filter to edit.
 * \param data Port to watch/update.
 *
 * \return 0 if everything goes well, -1 otherwise
 */
int filter_add_port(filter_table_t *filter, sdm_ip_t *ip_addr,
                    uint16_t port, uint32_t proto_flag, uint32_t rid)
{
    int ret_ins = FHF_INSERT_OK;
    int8_t *data_lock = NULL;
    void *updated = NULL;

    ret_ins = fhf_insert_own_or_update(filter->port_filter,
                                         ip_addr,
                                         &data_lock,
                                         &updated);

    if (ret_ins == FHF_INSERT_FAILED ||  ret_ins == FHF_INSERT_OK) {
        // Item can be inserted or is already in filter
        ((uint32_t *) updated)[port] = rid;
        ((uint32_t *) updated)[port] |= proto_flag;
        fhf_unlock_data(data_lock);
    }

    return ret_ins;
}
#endif
/**
 * \brief Procedure for removing prefix from filter.
 *
 * Procedure removes prefix filter rule from the filter table.
 * 
 * \param filter Filter for editing.
 * \param pref Removed prefix.
 * \param ip_ver Version of ip address.
 */
void filter_remove_prefix(filter_table_t *filter, 
                          pref_item_t *pref,
                          uint8_t ip_ver)
{
    if (ip_ver == 4) {
        pref_rem(&(filter->v4_pref), pref);
    } else {
        pref_rem(&(filter->v6_pref), pref);
    }
}

/**
 * \brief Procedure for removing flow from hash filter.
 *
 * Procedure removes flow filtering rule from the hash table.
 * 
 * \param filter Edited filter.
 * \param key Key of the removed flow.
 */
void filter_remove_flow(filter_table_t *filter, sdm_flow_identifier_t *key)
{
    fhf_remove(filter->flow_filter, key->raw);
}

/**
 * \brief Procedure for removing flow from hash filter.
 *
 * Procedure removes port filtering rule from the hash table.
 * 
 * \param filter Edited filter.
 * \param key Key of the removed flow.
 */
#ifndef NO_PORT
void filter_remove_port(filter_table_t *filter, sdm_ip_t *ip_addr,
                        uint16_t port, uint32_t proto_flag)
{
    void *removed = NULL;
    int in_table = FHF_NOT_FOUND;
    int remove_all = 1;
    int8_t *data_lock = NULL;
    
    in_table = fhf_update_data(filter->port_filter, ip_addr,
                               &data_lock, &removed);

    if (in_table == FHF_FOUND && removed != NULL) {
        ((uint32_t *) removed)[port] &= ~proto_flag;

        if (((PORT_TCP_FLAG | PORT_UDP_FLAG) & ((uint32_t *) removed)[port]) == 0)
            // port is no longer used --> invalidate RID
            ((uint32_t *) removed)[port] = INVALID_RID;
    

        for (uint32_t i = 0; i < UINT16_MAX + 1; i++) {
            // check other half-flows for ip address
            if (((uint32_t *) removed)[i] != INVALID_RID
                && ((uint32_t *) removed)[i] != 0x0) {
                remove_all = 0;
                break;
            }
        }

        fhf_unlock_data(data_lock);
    }
    if (remove_all == 1) {
        // no other ports monitored --> remove item from filter
        fhf_remove(filter->port_filter, ip_addr);
    }
}
#endif
/**
 * \brief Procedure for destroying filter table.
 *
 * Procedure deallocates all resources used by the filter table.
 * 
 * \param filter Filter to be disposed.
 */
void filter_destoy(filter_table_t *filter)
{
    pref_destroy(&(filter->v4_pref));
    pref_destroy(&(filter->v4_pref));

    fhf_destroy(filter->flow_filter);
#ifndef NO_PORT
    fhf_destroy(filter->port_filter);
#endif
}
