/**
 * \file filter_control.h
 * \brief Functions and procedures for filter table maintenace -- header file
 * \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
 * \date 2014
 */

#ifndef FILTER_CONTROL_H
#define FILTER_CONTROL_H

#include <stdint.h>
#include <endian.h>
#include "fast_hash_filter/fast_hash_filter.h"
#include "ccci_struct/ccci_rule.h"
#include "filter_types.h"
#include "pref_filter.h"

#include <sdm.h>

#define HASH_FILTER_SIZE 4096 // for generic use. will be removed in future if not used

/**
 * Size of flow filter hash table. MUST BE POWER OF 2
 */
#define FLOW_FILTER_SIZE 512 // size of flow filter
/**
 * Size of port filter hash tables. MUST BE POWER OF 2
 */
#define PORT_FILTER_SIZE 512

/* Send flags for items. */
#define SEND_ALL         0x03
#define SEND_SLIS        0x02
#define SEND_IRI_IIF     0x01
#define SEND_NONE        0x0

/* Port map protocol flags */
#define PORT_TCP_FLAG   0x00020000
#define PORT_UDP_FLAG   0x00040000

/* Match constants */
#define NO_MATCH        0
#define MATCH_FOUND     1

#define INVALID_RID     0x00010000
#define RID_VALUE(rid)  (rid & 0x0000FFFF)
/** 
 * Set of filter tables for each thread. 
 */
typedef struct filter_table {
    /*@{*/
    pref_filter_t v4_pref; /**< IPv4 prefix table. */
    pref_filter_t v6_pref; /**< IPv6 prefix table. */
    fhf_table_t *flow_filter; /**< Flow filter (8-way hash table). */
    fhf_table_t *port_filter; /**< Port filter (Array of small FHFs). */
    /*@}*/
} filter_table_t;

/* create key for hash filter */
void filter_key_ccci(union ccci_rule *rule, sdm_flow_identifier_t *key);

/* initialize filter tables */
int filter_init(filter_table_t *filter);

/* add item to prefix filter */
void filter_add_prefix(filter_table_t *filter,
                       pref_item_t *data,
                       uint8_t ip_ver);

/* add item to flow filter */
int filter_add_flow(filter_table_t *filter, filter_item_t *data);

/* add item to ip-port filter */
int filter_add_port(filter_table_t *filter, sdm_ip_t *ip_addr,
                    uint16_t port, uint32_t proto_flag, uint32_t rid);

/**
 * \brief Function for matching the data to any rule.
 *
 * Function takes the key created from data recieved via SZE interface and tries
 * to match it to any rule in the whole filter table. Function reports 
 * success or failure via corresponding structure match_t.
 * 
 * \param filter Filter table with rules.
 * \param key Key of the recieved frame.
 * \param ip_ver Version of ip addresses.
 * \param hs Set of precomputed hashes for hash filter.
 *
 * \return Structure match_t with report about success or failure of the search.
 */
static inline void filter_match(filter_table_t *filter,
                                sdm_flow_identifier_t *key,
                                match_t *m)
{

    const void *found = NULL;
    int fhf_state = FHF_NOT_FOUND;
    uint32_t rid_record = 0x0;

#ifdef DEBUG
    printf("Search flow key:");
    for (int i = 0; i < 40; i++)
        printf("%2x ", key->raw[i]);
    printf("\b\n");
#endif
    
    fhf_state = fhf_get_data(filter->flow_filter, key->raw, &found);
#ifdef DEBUG
    printf("Flow table entry: %p\n", found);
#endif
    if (fhf_state == FHF_FOUND) {
        m->match |= MATCH_FOUND;
        m->src_rid = m->dst_rid = *((uint32_t *) found);
        return;
    }



#ifndef NO_PORT
    fhf_state = fhf_get_data(filter->port_filter,
                             &(key->elems.src_ip),
                             &found);
    if (fhf_state == FHF_FOUND) {
        rid_record = ((uint32_t *) found)[key->elems.src_port];
        if (rid_record != INVALID_RID
            && ((key->elems.protocol == L4_TCP && rid_record & PORT_TCP_FLAG) 
            || (key->elems.protocol == L4_UDP && rid_record & PORT_UDP_FLAG))) 
        {
             m->match = MATCH_FOUND;
             m->src_rid = rid_record;
        }
    }

    fhf_state = fhf_get_data(filter->port_filter,
                             &(key->elems.dst_ip),
                             &found);

    if (fhf_state == FHF_FOUND) {
        rid_record = ((uint32_t *)found)[key->elems.dst_port];
        if (rid_record != INVALID_RID
            && ((key->elems.protocol == L4_TCP && rid_record & PORT_TCP_FLAG) 
            || (key->elems.protocol == L4_UDP && rid_record & PORT_UDP_FLAG))) 
        {
             m->match = MATCH_FOUND;
             m->dst_rid = rid_record;
        }
    }

#endif
    if (m->match == MATCH_FOUND) {
        if (m->src_rid == INVALID_RID) {
            m->src_rid = m->dst_rid;
        } else if (m->dst_rid == INVALID_RID) {
            m->dst_rid = m->src_rid;
        }
        return;
    }

#ifdef IPV6
    if (key->elems.ip_version == 4) {
#endif
        pref_match_v4(&(filter->v4_pref), &(key->elems.src_ip),
                      &(m->match), &(m->src_rid));
        pref_match_v4(&(filter->v4_pref), &(key->elems.dst_ip),
                      &(m->match), &(m->dst_rid));
#ifdef IPV6
    } else {
        pref_match_v6(&(filter->v6_pref), &(key->elems.src_ip),
                      &(m->match), &(m->src_rid));
        pref_match_v6(&(filter->v6_pref), &(key->elems.dst_ip), 
                      &(m->match), &(m->dst_rid));
    }
#endif

    if (m->match == MATCH_FOUND) {
        if (m->src_rid == INVALID_RID) {
            m->src_rid = m->dst_rid;
        } else if (m->dst_rid == INVALID_RID) {
            m->dst_rid = m->src_rid;
        }
        return;
    }

}

/* remove item from prefix filter */
void filter_remove_prefix(filter_table_t *filter,
                          pref_item_t *pref,
                          uint8_t ip_ver);

/* remove item from hash filter */
void filter_remove_flow(filter_table_t *filter,
                        sdm_flow_identifier_t *key);

/* remove port from filter */
void filter_remove_port(filter_table_t *filter, sdm_ip_t *ip_addr,
                        uint16_t port, uint32_t proto_flag);

/* destroy filter set */
void filter_destoy(filter_table_t *filter);

#endif  /* FILTER_CONTROL_H */
