#define _GNU_SOURCE
/*
 * File: filterd.c -- Filtering daemon for LI HW probe
 * Author: Roman Vrana, xvrana20@stud.fit.vutbr.cz
 * Date: 2013
 */  

#include <stdio.h>
#include <syslog.h>
#include <stdlib.h>
#include <stdint.h>
#include <signal.h>
#include <pthread.h>
#include <getopt.h>
#include <libtrap/trap.h>
#include <errno.h>

#ifdef PCAP_IFC
#include <pcap.h> // possibly deprecated
#else
#include <libsze2.h>
#endif

#include <sched.h>
#include <unistd.h>

#include "filter_types.h"
#include "filterd.h"
#include "inet_offsets.h"
#include "iri_iif_req.h"
#include "../ini3d/ini3d.h"
#include "ccci_struct/ccci_msg.h"
#include "ccci_struct/ccci_rule.h"
#include "../libs/error_logger/error_logger.h"

#include "iri_iif/controller.h"
#include "iri_iif/adb.h"

#define SDM_INSTALLED
#define SZE_PATH "/dev/szedataII0"

#ifdef SDM_INSTALLED
#include <sdm.h>
#else
#include "libsdm/sdm.h"
#endif

//#define NO_HW

#define THR_MAX 8

#define EXT_HEADERS_INIT(e) e[0] = 1; e[43] = 1; e[44] = 1; e[50] = 1; \
                            e[51] = 1; e[60] = 1; e[135] = 1; e[139] = 1; \
                            e[140] = 1

static int iri_whitelist[65536];
#define IRI_WHITELIST_INIT() memset(iri_whitelist, 0x0, 65536); \
                             iri_whitelist[5060] = 1;\
                             iri_whitelist[5062] = 1;\
                             iri_whitelist[25] = 1;\
                             iri_whitelist[587] = 1;


// TRAP signal handling
static int stop = 0;

static int thr_count = THR_MAX;

TRAP_DEFAULT_SIGNAL_HANDLER(stop = 1);

trap_module_info_t module_info = {
    "Filter daemon for CC-IIF block of the Sec6Net LI system",
    "Module for filtering traffic from SZE interface and controlling the input\n"
    "via SDM.\n",
    1, // CCCI request message from CCCId
    1, // INI3 payload
};

static logger_t *error_log = NULL;

#ifdef DEMO
#define SDM_CONTEXT_ID 0
#endif

/**
 * Function for getting frame type and setting correct offsets for frame 
 * parsing. This function shares the combines the task of previous functions
 * get_ether_type and filter_key_frame. Function puts all values in LE byte 
 * order so it corresponds with SDM rule specification. If you need to keep 
 * network you have to swap these values again.
 *
 * \param [in] frame Pointer to L2 frame.
 * \param [out] offsets Structure with offsets for required fields.
 * \param [out] l_key Filled in lookup key for filtering table.
 */
static inline void parse_frame(unsigned char *frame,
                               offsets_t *offsets,
                               sdm_flow_identifier_t *l_key,
                               uint8_t* v6_ext_headers)
{
    unsigned int vlan = 0;
    uint32_t *ip;
    uint16_t *port;

    offsets->iph_start = ETHER_LENGTH;
    // is 802.1Q ? 
    if (ntohs(*((uint16_t *) (frame + ETHER_TYPE_OFF))) == ETH_VLAN) {
        offsets->iph_start += VLAN_TAG;
        vlan = VLAN_TAG; 
    }

    if (ntohs(*((uint16_t *) (frame + vlan + ETHER_TYPE_OFF))) == ETH_IP4) {

        // IPv4 frame
        offsets->ip_ver = l_key->elems.ip_version = 4;
        offsets->ip_start = offsets->iph_start + IP4_START;
        ip = (uint32_t *) (frame + offsets->ip_start);

        int ihl = (*(frame + offsets->iph_start) << 4) >> 2;
        offsets->l4_start = offsets->iph_start + ihl;
        offsets->protocol = offsets->iph_start + IP4_PROTOCOL;
        port = (uint16_t *) (frame + offsets->l4_start);

        // fill in key in LE byte order
        if (ntohl(*(ip)) <= ntohl(*(ip + 1))) {
    	    l_key->elems.src_ip._v4.d32[0] = ntohl(*(ip));
            l_key->elems.dst_ip._v4.d32[0] = ntohl(*(ip + 1));
            l_key->elems.src_port = ntohs(*(port));
            l_key->elems.dst_port = ntohs(*(port + 1));
        } else {
	        l_key->elems.src_ip._v4.d32[0] = ntohl(*(ip + 1));
            l_key->elems.dst_ip._v4.d32[0] = ntohl(*(ip));
            l_key->elems.src_port = ntohs(*(port + 1));
            l_key->elems.dst_port = ntohs(*(port));
        }

        l_key->elems.protocol = *(frame + offsets->protocol);
#ifdef IPV6
    } else if (ntohs(*((uint16_t *) (frame + vlan + ETHER_TYPE_OFF))) == ETH_IP6) {
        // IPv6 frame
        offsets->ip_ver = l_key->elems.ip_version = 6;
        offsets->ip_start = offsets->iph_start + IP6_START;
        ip = (uint32_t *) (frame + offsets->ip_start);
        offsets->l4_start = offsets->iph_start + IP6_H_LENGTH;
        port = (uint16_t *) (frame + offsets->l4_start);
        offsets->protocol = offsets->iph_start + IP6_PROTOCOL;
        
#ifdef EXT_HEADERS_SKIP
        int ext_hdr_len;
        uint8_t v6_proto = *(frame + offsets->protocol);
        if (v6_ext_headers[v6_proto] == 1) {
            do {               
                offsets->protocol = offsets->l4_start;
                ext_hdr_len = (port & 0x00FF) >> 1;
                v6_proto = *(frame + offsets->protocol);
                offsets->l4_start += ext_hdr_len;
                port += ext_hdr_len;
            } while (v6_ext_headers[offsets->protocol] == 1);
        }
#endif

        l_key->elems.src_ip.v6.d32[3] = ntohl(*(ip));
        l_key->elems.src_ip.v6.d32[2] = ntohl(*(ip + 1));
        l_key->elems.src_ip.v6.d32[1] = ntohl(*(ip + 2));
        l_key->elems.src_ip.v6.d32[0] = ntohl(*(ip + 3));

        l_key->elems.dst_ip.v6.d32[3] = ntohl(*(ip + 4));
        l_key->elems.dst_ip.v6.d32[2] = ntohl(*(ip + 5));
        l_key->elems.dst_ip.v6.d32[1] = ntohl(*(ip + 6));
        l_key->elems.dst_ip.v6.d32[0] = ntohl(*(ip + 7));

        l_key->elems.src_port = ntohs(*(port));
        l_key->elems.dst_port = ntohs(*(port + 1));

        l_key->elems.protocol = *(frame + offsets->protocol);
#endif
    } else {
        // unknown protocol skip frame
        offsets->ip_ver = 0;
    }
}
/**
 * Function for adding new interception based on give network prefix.
 * \param [in] msg CCCI message containing rule.
 * \param [in] filter_control Pointers to filtering tables.
 */
int new_interception_prefix(union ccci_msg *msg, manager_args_t *filter_control, intercept_identifier_t *iri_ict)
{
    union ccci_rule *rule = ccci_msg_getrule(msg);
    unsigned ip_ver = ccci_rule_getipver(rule);
    pref_item_t *new_pref = (pref_item_t*) calloc(1,sizeof(pref_item_t));
    if (new_pref == NULL) {
        log_msg(error_log, 
                "Unable to allocate new item. Terminating ...\n",
                LOG_ERR);
        return 1;
    }

#ifdef DEBUG
    printf("IP: %hhu.%hhu.%hhu.%hhu\nRID: %u\n", 
            ccci_rule_getip0(rule)->raw[3],
            ccci_rule_getip0(rule)->raw[2],
            ccci_rule_getip0(rule)->raw[1],
            ccci_rule_getip0(rule)->raw[0],
            ccci_msg_getrid(msg)
          );
#endif

    new_pref->rid = ccci_msg_getrid(msg);
    iri_ict->rid = ccci_msg_getrid(msg);

    if (ip_ver == 4) {
        new_pref->prefix._v4.d32[0] = ccci_rule_getip0(rule)->raw32[0];
        new_pref->length = ccci_rule_getipmask(rule);
        iri_ict->key.items.ip_1._v4.d32[0] = ccci_rule_getip0(rule)->raw32[0];
    } else {
        memcpy(&(new_pref->prefix), ccci_rule_getip0(rule), sizeof(sdm_ip_t));
        new_pref->length = ccci_rule_getipmask(rule);
        memcpy(&(iri_ict->key.items.ip_1.v6), ccci_rule_getip0(rule), sizeof(sdm_ip_t));
    }

    // add to filter for all threads
    for (int i = 0; i < thr_count; i++) 
        filter_add_prefix(&(filter_control->filter[i]), new_pref, ccci_rule_getipver(rule));

    return 0;
}

#ifndef NO_HW
int new_interception_flow(union ccci_msg *msg,
                          manager_args_t *filter_control,
                          sdm_message_t *sdm_rule, intercept_identifier_t *int_id)
#else
int new_interception_flow(union ccci_msg *msg, manager_args_t *filter_control, intercept_identifier_t *int_id)
#endif
{
    union ccci_rule *rule = ccci_msg_getrule(msg);
    filter_item_t *new_item = (filter_item_t*) calloc(1,sizeof(filter_item_t));

    if (new_item == NULL) {
        log_msg(error_log, 
                "Unable to allocate new item. Terminating ...\n",
                LOG_ERR);
        trap_terminate(); 
        return 1;
    }

    new_item->rid = ccci_msg_getrid(msg);
    filter_key_ccci(rule, &(new_item->key));

    // add to filter for all threads
#ifdef DEBUG
    log_msg(error_log, "Inserting flow rule to filter.\n", LOG_DEBUG);
#endif
    for (int i = 0; i < thr_count; i++) {
        int ins_state = filter_add_flow(&(filter_control->filter[i]), new_item);
        switch (ins_state) {
            // TODO: format log strings
            case FHF_INSERT_FULL:
                log_msg(error_log,
                        "Unable to add flow to table. Table is full.\n",
                        LOG_WARNING);
                break;

            case FHF_INSERT_FAILED:
                log_msg(error_log,
                        "Unable to add flow to table. Flow filter error.\n",
                        LOG_ERR);
                break;
        }
    }
#ifndef NO_HW
    // prepare SDM key (cannot be done outside due the access to items)
    sdm_message_access_key(sdm_rule) = new_item->key;
    sdm_flow_wildcard_empty(&(sdm_rule->key));
    sdm_flow_wildcard_set_src_ip_mask(&(sdm_rule->key), new_item->key.elems.ip_version == 4 ? 32 : 128);
    sdm_flow_wildcard_set_dst_ip_mask(&(sdm_rule->key), new_item->key.elems.ip_version == 4 ? 32 : 128);
    sdm_flow_wildcard_set_src_port(&(sdm_rule->key));
    sdm_flow_wildcard_set_dst_port(&(sdm_rule->key));
    sdm_flow_wildcard_set_protocol(&(sdm_rule->key));
    sdm_flow_wildcard_set_ip_version(&(sdm_rule->key));
#endif
    memcpy(&(int_id->key), new_item, 40);
    free(new_item);
    return 0;
}

void new_interception_port(union ccci_msg *msg, manager_args_t *filter_control)
{
    union ccci_rule *rule = ccci_msg_getrule(msg);
    unsigned ip_ver = ccci_rule_getipver(rule);
    sdm_ip_t ip_addr;
    memset(&ip_addr, 0x0, sizeof(sdm_ip_t));

    if (ip_ver == 4) {
        ip_addr.v4_long.data.d32[0] = ccci_rule_getip0(rule)->raw32[0];
    } else {
        ip_addr.v6.d64[0] = ccci_rule_getip0(rule)->raw64[0];
        ip_addr.v6.d64[1] = ccci_rule_getip0(rule)->raw64[1];
    }

    // add to filter for all threads
    for (int i = 0; i < thr_count; i++) 
        filter_add_port(&(filter_control->filter[i]),
                          &ip_addr,
                          ccci_rule_getport0(rule),
                          ccci_rule_getproto4(rule) == L4_TCP ? PORT_TCP_FLAG : PORT_UDP_FLAG,
                          ccci_msg_getrid(msg));
}

void del_interception_prefix(union ccci_msg *msg, manager_args_t *filter_control)
{
    union ccci_rule *rule = ccci_msg_getrule(msg);
    pref_item_t rem_pref;
    unsigned ip_ver = ccci_rule_getipver(rule);
    if (ip_ver == 4) {
        rem_pref.prefix._v4.d32[0] = ccci_rule_getip0(rule)->raw32[0];
        rem_pref.length = ccci_rule_getipmask(rule);
    } else {
        memcpy(&(rem_pref.prefix), ccci_rule_getip0(rule), sizeof(sdm_ip_t));
        rem_pref.length = ccci_rule_getipmask(rule);
    }
    
    // remove from filter for all threads
    for (int i = 0; i < thr_count; i++)
        filter_remove_prefix(&(filter_control->filter[i]),
                               &rem_pref,
                               ccci_rule_getipver(rule));

}

void del_interception_flow(union ccci_msg *msg, manager_args_t *filter_control)
{
    sdm_flow_identifier_t rem_key;
    memset(&rem_key, 0x0, sizeof(rem_key));
    filter_key_ccci(ccci_msg_getrule(msg), &rem_key);

    // remove from filter for all threads
    for (int i = 0; i < thr_count; i++)
        filter_remove_flow(&(filter_control->filter[i]), &(rem_key));
}

void del_interception_port(union ccci_msg *msg, manager_args_t *filter_control)
{
    union ccci_rule *rule = ccci_msg_getrule(msg);
    unsigned ip_ver = ccci_rule_getipver(rule);
    sdm_ip_t ip_addr;
    memset(&ip_addr, 0x0, sizeof(sdm_ip_t));

    if (ip_ver == 4) {
        ip_addr.v4_long.data.d32[0] = ccci_rule_getip0(rule)->raw32[0];
    } else {
        ip_addr.v6.d64[0] = ccci_rule_getip0(rule)->raw64[0];
        ip_addr.v6.d64[1] = ccci_rule_getip0(rule)->raw64[1];
    }

    // remove from filter for all threads
    for (int i = 0; i < thr_count; i++) 
        filter_remove_port(&(filter_control->filter[i]),
                          &ip_addr,
                          ccci_rule_getport0(rule),
                          ccci_rule_getproto4(rule) == L4_TCP ? PORT_TCP_FLAG : PORT_UDP_FLAG);
}

/**
 * Function for managing filters
 */
void* filter_manager(void *args)
{
    manager_args_t *man_args = (manager_args_t *) args; 
    
    const void *li_req;
    uint16_t data_size;
    int trap_ret;

#ifndef NO_HW
    /* *** SDM INIT *** */

    sdm_t sdm_control;
    sdm_access_t sdm_acc;

    sdm_control = sdm_init();
    if (sdm_control == NULL) {
        log_msg(error_log,
                "Unable to initialize SDM. Terminating...\n",
                LOG_ERR);
        stop = 1;
        trap_terminate();
        return 0;
    }
        
    sdm_acc = sdm_open(sdm_control, SDM_BUFFER_SIZE);
    if (sdm_acc == NULL) {
        log_msg(error_log,
                "Unable to open SDM communication. Terminating...\n",
                LOG_ERR);
        stop = 1;
        trap_terminate();
        return 0;
    }

    sdm_message_t *rule_set;
#ifndef DEMO
    int sdm_ctx_id;
    context_t sdm_ctx;
    memset(&sdm_ctx, 0x0, sizeof(context_t));
    sdm_ctx.min_channel = 0;
    sdm_ctx.max_channel = 7;
    sdm_ctx.default_action = SDM_ACTION_FULL;
    sdm_ctx.rules = 1;

    sdm_ctx_id = sdm_start_context(sdm_control, &sdm_ctx);
    printf("%d\n", sdm_ctx_id);
    if (sdm_ctx_id == -1) {
       log_msg(error_log,
               "SDM context not set. Requests will not be checked.\n",
               LOG_WARNING);
    }
#endif
    /* *** SDM INIT *** */
#endif
    int rule_type;
    int action = NO_INTERCEPT;

    while (!stop) {
        action = NO_INTERCEPT;

        // recieve message from CCCID
        trap_ret = trap_recv(0, &li_req, &data_size);
        TRAP_DEFAULT_RECV_ERROR_HANDLING(trap_ret, ;, break);

        if (trap_ret != TRAP_E_TIMEOUT) {
            rule_type = ccci_rule_gettype(ccci_msg_getrule((union ccci_msg *) li_req));
            action = ccci_msg_getaction(li_req);
        }            
        

        /* serve request from CCCId */
        if (action == INTERCEPT_ADD) {
#ifdef DEBUG
        log_msg(error_log, "Parsing message from CCCI.\n", LOG_DEBUG);
#endif
            if (rule_type == 0) {
                intercept_identifier_t iri_ict;
                memset(&iri_ict.key, 0x0, 40);
                new_interception_prefix((union ccci_msg *) li_req, man_args, &iri_ict);
#ifndef NO_HW
                union ccci_rule *rule = ccci_msg_getrule((union ccci_msg *) li_req);
                rule_set = sdm_get_message_buffer(sdm_acc);
                memcpy(&(sdm_message_access_key(rule_set).elems.src_ip),
                       ccci_rule_getip0(rule), 16);
                sdm_message_access_key(rule_set).elems.ip_version = ccci_rule_getipver(rule);

                sdm_flow_wildcard_empty(&(rule_set->key));
                sdm_flow_wildcard_set_src_ip_mask(&(rule_set->key), ccci_rule_getipmask(rule));
                sdm_flow_wildcard_set_ip_version(&(rule_set->key));
                sdm_message_set_operation(rule_set, SDM_MESSAGE_OPERATION_DELETE_GROUP);

                while(sdm_send(sdm_acc) == 0);

                rule_set = sdm_get_message_buffer(sdm_acc);
                memcpy(&(sdm_message_access_key(rule_set).elems.dst_ip),
                       ccci_rule_getip0(rule), 16);
                sdm_message_access_key(rule_set).elems.ip_version = ccci_rule_getipver(rule);

                sdm_flow_wildcard_empty(&(rule_set->key));
                sdm_flow_wildcard_set_dst_ip_mask(&(rule_set->key), ccci_rule_getipmask(rule));
                sdm_flow_wildcard_set_ip_version(&(rule_set->key));
                sdm_message_set_operation(rule_set, SDM_MESSAGE_OPERATION_DELETE_GROUP);

                while(sdm_send(sdm_acc) == 0);
#endif
            } else if (rule_type == 1) {
                intercept_identifier_t iri_ict;
#ifndef NO_HW
                rule_set = sdm_get_message_buffer(sdm_acc);

                new_interception_flow((union ccci_msg *) li_req, man_args, rule_set, &iri_ict);

                memcpy(&iri_ict, rule_set, 40);
                iri_ict.rid = ccci_msg_getrid((union ccci_msg *) li_req);
#ifdef DEBUG
                char flow_id_log[1024];
                sprintf(flow_id_log,
                        "%s %d: Sending added flow log to IRI-IIF %s\n",
                        __FILE__,
                        __LINE__,
                        (char *) &iri_ict);
                log_msg(error_log, flow_id_log, LOG_DEBUG);
                memset(flow_id_log, 0x0, 1024);
#endif
                itc_write(man_args->interface_to_controller, 0, (char *) &iri_ict);

                sdm_message_set_operation(rule_set, SDM_MESSAGE_OPERATION_DELETE_GROUP);

                
                // send rule to HW
                while(sdm_send(sdm_acc) == 0);

#else
                new_interception_flow((union ccci_msg *) li_req, man_args, &iri_ict);
                iri_ict.rid = ccci_msg_getrid((union ccci_msg *) li_req);
#ifdef DEBUG
                char flow_id_log[1024];
                sprintf(flow_id_log,
                        "%s %d: Sending added flow log to IRI-IIF %s\n",
                        __FILE__,
                        __LINE__,
                        (char *) &iri_ict);
                log_msg(error_log, flow_id_log, LOG_DEBUG);
                memset(flow_id_log, 0x0, 1024);
#endif

                itc_write(man_args->interface_to_controller, 0, (char *) &iri_ict);

#endif /* NO_HW */
            } else if (rule_type == 3) {
                new_interception_port((union ccci_msg *) li_req, man_args);
#ifndef NO_HW
                union ccci_rule *rule = ccci_msg_getrule((union ccci_msg *) li_req);
                rule_set = sdm_get_message_buffer(sdm_acc);
                memcpy(&(sdm_message_access_key(rule_set).elems.src_ip),
                       ccci_rule_getip0(rule), 16);
                sdm_message_access_key(rule_set).elems.src_port = ccci_rule_getport0(rule);
                sdm_message_access_key(rule_set).elems.protocol = ccci_rule_getproto4(rule);
                sdm_message_access_key(rule_set).elems.ip_version = ccci_rule_getipver(rule);

                sdm_flow_wildcard_empty(&(rule_set->key));
                sdm_flow_wildcard_set_src_ip_mask(&(rule_set->key), ccci_rule_getipmask(rule));
                sdm_flow_wildcard_set_src_port(&(rule_set->key));
                sdm_flow_wildcard_set_protocol(&(rule_set->key));
                sdm_flow_wildcard_set_ip_version(&(rule_set->key));

                sdm_message_set_operation(rule_set, SDM_MESSAGE_OPERATION_DELETE_GROUP);

                while(sdm_send(sdm_acc) == 0);

                rule_set = sdm_get_message_buffer(sdm_acc);
                memcpy(&(sdm_message_access_key(rule_set).elems.dst_ip),
                       ccci_rule_getip0(rule), 16);
                sdm_message_access_key(rule_set).elems.dst_port = ccci_rule_getport0(rule);
                sdm_message_access_key(rule_set).elems.protocol = ccci_rule_getproto4(rule);
                sdm_message_access_key(rule_set).elems.ip_version = ccci_rule_getipver(rule);

                sdm_flow_wildcard_empty(&(rule_set->key));
                sdm_flow_wildcard_set_dst_ip_mask(&(rule_set->key), ccci_rule_getipmask(rule));
                sdm_flow_wildcard_set_src_port(&(rule_set->key));
                sdm_flow_wildcard_set_protocol(&(rule_set->key));
                sdm_flow_wildcard_set_ip_version(&(rule_set->key));

                sdm_message_set_operation(rule_set, SDM_MESSAGE_OPERATION_DELETE_GROUP);

                while(sdm_send(sdm_acc) == 0);
#endif
            }

        } else if (action == INTERCEPT_REMOVE) {
            if (rule_type == 0) {
                del_interception_prefix((union ccci_msg *) li_req, man_args);
            } else if (rule_type == 1) {
                del_interception_flow((union ccci_msg *) li_req, man_args);
            } else if (rule_type == 3) {
                del_interception_port((union ccci_msg *) li_req, man_args);
            }
        }



    // check SDM request queue
#ifndef NO_HW
        for (int i = 0; i < thr_count; i++) {
            for (int q = 0; q < SDM_BATCH_SIZE; q++) {            
                // pull from sdm queue
                
                sdm_flow_identifier_t sdm_key = sdm_q_pull(&(man_args->sdm_req[i]));
                if (sdm_key.raw[0] == 0x1)
                    break;                

                // create sdm key
                rule_set = sdm_get_message_buffer(sdm_acc);
                memset(rule_set, 0x0, sizeof(sdm_message_t));
                sdm_message_access_key(rule_set) = sdm_key;


                sdm_message_set_action(rule_set, SDM_ACTION_DROP);
                sdm_message_set_operation(rule_set, SDM_MESSAGE_OPERATION_ADD);
#ifdef DEMO
                sdm_message_set_context(rule_set, SDM_CONTEXT_ID);
#else
                sdm_message_set_context(rule_set, sdm_ctx_id);
#endif
                sdm_message_add_operation(rule_set, SDM_MESSAGE_OPERATION_MARKHEAVY); 
                sdm_message_set_nohash(rule_set);

                // send rule to SDM controller
                sdm_send(sdm_acc);
            }
        }
#endif
    }
#ifndef NO_HW

#ifndef DEMO
    sdm_end_context(sdm_control, sdm_ctx_id);
#endif

    sdm_deinit(sdm_control);
#endif
    return NULL;
}


/**
 * Thread for filtering packets from SZE interface
 */
void* filter_report(void *args)
{
    report_args_t *filter = (report_args_t *) args;
    int trap_ret;
    unsigned char send_buffer[1600];

#ifdef PCAP_IFC // PCAP interface for packet capture
    pcap_t *pcap_handle = NULL;
    char pcap_err[PCAP_ERRBUF_SIZE];


    int ret = 0;

    pcap_handle = pcap_open_live(filter->pcap_dev, BUFSIZ, 1, 0, pcap_err);

    // PCAP open failure
    if (pcap_handle == NULL) {
        log_msg(error_log,
                "Cannot open Pcap interface.\n",
                LOG_ERR);
        trap_terminate();
        ret = -1;
        return (void *) ret;
    }

    struct pcap_pkthdr pcap_head; // pcap capture header
#else // SZE_IFC 
    struct szedata *sze_handle = NULL;
    
    // path to SZE device should be same on every machine
    const char *sze_path = SZE_PATH;

    unsigned rx, tx = 0x0; // mask of RX (read) and TX (write) channels
    
    int ret = 0;

    sze_handle = szedata_open(sze_path);

    // SZE open failure
    if (sze_handle == NULL) {
        log_msg(error_log,
                "Cannot open SZE interface.\n",
                LOG_ERR);
        trap_terminate();
        ret = -1;
        return (void *) ret;
    }

    // lock thread to core
    cpu_set_t cpus;
    CPU_ZERO(&cpus);
    CPU_SET(filter->thr_id, &cpus);

    pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t), &cpus);


    // assign RX channels for multithreaded filtering
    switch(thr_count) {
        case 1:
            rx = 0xFF;
            break;
        case 2:
            rx = 0xF;
            break;
        case 4:
            rx = 0x3;
            break;
        case 8:
            rx = 0x1;
            break;
        default:
            rx = 0xFF;
            break;
    }

    rx <<= (filter->thr_id * ((sizeof(char)*8) / thr_count));

    // SZE interface polling timeout (0 for active wait polling)
    // If active waiting is an issue remove the line from code
    SZE2_RX_POLL_TIMEOUT = 1;

    // subscribe/bind channels of interface
    ret = szedata_subscribe3(sze_handle, &rx, &tx);
    if (ret != 0) {
        log_msg(error_log,
                "Cannot subscribe to SZE interface\n",
                LOG_ERR);
        trap_terminate();
        ret = -1;
        return (void *) ret;
    }

    // open capture
    ret = szedata_start(sze_handle);
    if (ret != 0) {
        log_msg(error_log,
                "Cannot open data channel to SZE interface\n",
		LOG_ERR);
        trap_terminate();
        ret = -1;
        return (void *) ret;
    }

    unsigned int frm_length; // length of whole frame
    unsigned char *frame_data, *hw_data; // pointer to SW data and HW data of frame
    unsigned int frame_data_length, hw_data_length; // Lengths of SW and HW data
    unsigned int segm_size;
#endif

    unsigned char* frame;

    /* 
     * allocate lookup key for filtering table 
     * heap allocation used to override packed size so filtering table can use
     * faster hash computation
     */
    sdm_flow_identifier_t *lookup_key;
    lookup_key = (sdm_flow_identifier_t *) calloc(1, sizeof(sdm_flow_identifier_t) + 1);
    
    offsets_t offsets;
    memset(&offsets, 0x0, sizeof(offsets_t));

//    uint32_t flow_hash;

    static match_t m;
    static int iri_result;

    int frame_type = 0;

    // Connect to IRI-IIF interface and create key of the network flow
    iri_iif_init(HASH_FILTER_SIZE, 
                 filter->iri_iif_socket,
                 filter->interface_for_controller,
                 filter->interface_for_adb,
                 filter->thr_id);
    flow_identifier_t fi;
    fi.key = (item_key_t *) calloc(1, sizeof(item_key_t));
    if (fi.key == NULL) {
        stop = 1;
    }

    ini3payload_t *report = calloc(1, sizeof(ini3payload_t));

    /* Do the job */
    while (!stop) {
        m.match = NO_MATCH;
        m.src_rid = m.dst_rid = INVALID_RID;
        iri_result = BLOCK_FLOW;

#ifdef PCAP_IFC
        frame = pcap_next(pcap_handle, &pcap_head);
#else
        frame = szedata_read_next(sze_handle, &frm_length);
#endif
        if (frame != NULL) {


#ifdef PCAP_IFC
            parse_frame(frame, &offsets, lookup_key, filter->ext_headers);
#else // SZE IFC
            // decode captured packet from SZE interface to respective structures
            segm_size = szedata_decode_packet(frame, &frame_data, &hw_data,
                                              &frame_data_length,
                                              &hw_data_length);

            offsets.frame_data = SZE_FRAME_OFFSET;
            parse_frame(frame_data, &offsets, lookup_key, filter->ext_headers);
#endif

            // extract ID parts (IPs, ports, etc.)
#ifdef DEBUG
            log_msg(error_log, "Matching to filter...\n", LOG_DEBUG);
#endif

            filter_match(filter->filter, lookup_key, &m);

            // create INI3 message
#ifdef PCAP_IFC
            report->header.blob_size = pcap_head.len;
            report->header.sze_int = 0x0;
            report->header.timestamp_sec = pcap_head.ts.tv_sec;
            report->header.timestamp_ns = 0x0;
            report->header.rid[0] = RID_VALUE(m.src_rid);
            report->header.rid[1] = RID_VALUE(m.dst_rid);
            report->data = frame;
#else
            report->parts.header.blob_size = frame_data_length;
            report->parts.header.sze_int = lookup_key->elems.interface = *(hw_data + 1);
            report->parts.header.timestamp_sec = *(((uint32_t *) hw_data) + 2);
            report->parts.header.timestamp_ns = *(((uint32_t *) hw_data) + 1);
            report->parts.header.rid[0] = m.src_rid;
            report->parts.header.rid[1] = m.dst_rid;
#endif

            // fill in flow key for IRI-IIF
            fi.key->items.ip_1 = lookup_key->elems.src_ip;
            fi.key->items.ip_2 = lookup_key->elems.dst_ip;
            fi.key->items.port_1 = lookup_key->elems.src_port;
            fi.key->items.port_2 = lookup_key->elems.dst_port;
            fi.key->items.proto = lookup_key->elems.protocol;
            fi.key->items.ipver = lookup_key->elems.ip_version;

            memcpy(report->parts.data, frame_data, report->parts.header.blob_size);
            if (iri_whitelist[lookup_key->elems.src_port] || iri_whitelist[lookup_key->elems.dst_port])
                iri_result = iri_iif_process_new_packet(&fi, &offsets, report);

            if (m.match != NO_MATCH) {
#ifdef DEBUG
                log_msg(error_log, "Match found -- sending packet ...\n", LOG_DEBUG);
#endif /* DEBUG */

#ifdef DEBUG
                printf("Created INI3 report:\n");
                printf("Size: %hu\nSrcRID: %hu\nDstRID: %hu\n",
                report->parts.header.blob_size, report->parts.header.rid[0], report->parts.header.rid[1]);
#endif
                // send to INI3d
                trap_ret = trap_send(0, report->raw, sizeof(report->parts.header) + report->parts.header.blob_size);
                TRAP_DEFAULT_SEND_ERROR_HANDLING(trap_ret, 0, 0);

                memset(report->parts.data, 0x0, 1600);

            } else if (m.match == NO_MATCH && iri_result == BLOCK_FLOW) {
#ifndef NO_HW
#ifdef DEBUG
                log_msg(error_log,
                        "No match. Preparing SDM message ...\n",
                        LOG_DEBUG);
                if (sdm_q_push(filter->sdm_req, lookup_key) == -1) {
                    log_msg(error_log,
                            "SDM Queue full. Request dropped.\n",
                            LOG_WARNING);
                } else {
                    log_msg(error_log, "SDM request accepted. \n", LOG_DEBUG);
                }
#else
                
                if (iri_whitelist[lookup_key->elems.src_port] == 0 
                    && iri_whitelist[lookup_key->elems.dst_port] == 0)
                    sdm_q_push(filter->sdm_req, lookup_key);
#endif /* DEBUG */
#endif /* NO_HW */
            }
            frame = NULL;
            memset(lookup_key, 0x0, sizeof(sdm_flow_identifier_t));
            memset(fi.key, 0x0, sizeof(item_key_t));
        }
    }
#ifdef PCAP_IFC
    pcap_close(pcap_handle);
#else
    szedata_close(sze_handle);
#endif /* PCAP_IFC */
    return NULL;
}

int main(int argc, char** argv)
{
    // init trap interfaces
    TRAP_DEFAULT_INITIALIZATION(argc, argv, module_info);
    TRAP_REGISTER_DEFAULT_SIGNAL_HANDLER();

#ifndef NO_HW
    trap_ifcctl(TRAPIFC_INPUT,0,TRAPCTL_SETTIMEOUT, TRAP_IN_TIMEOUT);
    trap_ifcctl(TRAPIFC_INPUT,0,TRAPCTL_BUFFERSWITCH,1);
#endif

    trap_ifcctl(TRAPIFC_OUTPUT,0,TRAPCTL_SETTIMEOUT,TRAP_NO_WAIT);
    trap_ifcctl(TRAPIFC_OUTPUT,0,TRAPCTL_BUFFERSWITCH,1);

    // getopt for additional parameters besides TRAP interface    
    int o = 0;
    char *dev_str = NULL;
    char *log_path = NULL;
    while ((o = getopt(argc, argv, "d:l:t:")) != -1) {
        switch(o) {
            case 'l':
                log_path = optarg;
                break;
#ifdef PCAP_IFC
            case 'd':
                dev_str = optarg;
                break;
#endif
            case 't':
                thr_count = atoi(optarg);
                break;

            default:
                break;
        }
    }

#ifdef DEBUG
    error_log = logger_init(LOG_STDERR, "FILTERD", NULL);
#else
    if (log_path == NULL) {
        error_log = logger_init(LOG_SLOG, "FILTERD", NULL);
    } else {
        error_log = logger_init(LOG_FILE, "FILTERD", log_path);
    }
#endif

#ifdef PCAP_IFC
    if (dev_str == NULL) {
        log_msg(error_log,
                "Missing mandatory argument -d: PCAP capture interface\n",
                LOG_ERR);
        logger_destroy(error_log);;
        return EXIT_FAILURE;
    }
#endif

    pthread_t filterd_manager, *filterd_report;
    static manager_args_t man_thr;
    static report_args_t *rep_thr; 
    int man_ret, *filt_ret; 
    static filter_table_t *filters;
    static sdm_q_t *sdm_requests;

    pthread_t iri_adb;
    iri_iif_adb_args_t *adb_thr;

    filterd_report = (pthread_t *) calloc(thr_count, sizeof(pthread_t));
    rep_thr = (report_args_t *) calloc(thr_count, sizeof(report_args_t));
    filt_ret = (int *) calloc(thr_count, sizeof(int));
    filters = (filter_table_t *) calloc(thr_count, sizeof(filter_table_t));
    sdm_requests = (sdm_q_t *) calloc(thr_count, sizeof(sdm_q_t));

    int iri_iif_socket = iri_iif_modules_open_socket_to_iri_core();
    
    itc_ptr interface_for_adb, interface_to_controller;
    if( itc_init(&interface_for_adb, thr_count, ITC_RECORDS_PER_CIRCLE, BYTES_PER_RECORD))
    {
        itc_destroy(&interface_for_adb);
        return EXIT_FAILURE;
    }
    if( itc_init(&interface_to_controller, thr_count, ITC_RECORDS_PER_CIRCLE, BYTES_PER_RECORD))
    {
        itc_destroy(&interface_to_controller);
        return EXIT_FAILURE;
    }

    IRI_WHITELIST_INIT();

    // initialize ext_headers array for frame parser
    static uint8_t ext_headers[UINT8_MAX + 1];
    memset(ext_headers, 0x0, UINT8_MAX + 1);
    EXT_HEADERS_INIT(ext_headers);

    // create filters and link filters (number specified by parameter?)
    int err = 0;
#ifdef DEBUG
        log_msg(error_log, "Initializing filters ...\n", LOG_DEBUG);
#endif



    for (int i = 0; i < thr_count; i++) {
        err = filter_init(&filters[i]);
        if (err) {
            // unable to init filter --> terminate
            log_msg(error_log,
                    "Cannot initialize filters. Terminating ...\n",
                    LOG_ERR);
            TRAP_DEFAULT_FINALIZATION();
            for (int d = 0; d < i; d++) {
                filter_destoy(&filters[d]);
            }
            return EXIT_FAILURE;
        }

        sdm_q_init(&(sdm_requests[i]));

        rep_thr[i].filter = &(filters[i]);
        rep_thr[i].thr_id = i;
        rep_thr[i].sdm_req = &(sdm_requests[i]);
#ifdef PCAP_IFC
        rep_thr[i].pcap_dev = dev_str;
#endif
        rep_thr[i].iri_iif_socket = iri_iif_socket;
        rep_thr[i].interface_for_controller = interface_to_controller;
        rep_thr[i].interface_for_adb = interface_for_adb;
        rep_thr[i].ext_headers = ext_headers;
    }
    man_thr.filter = filters;
    man_thr.reporter_count = thr_count;
    man_thr.sdm_req = sdm_requests;

    adb_thr = (iri_iif_adb_args_t *) malloc(sizeof(iri_iif_adb_args_t));
    adb_thr->iri_iif_socket = iri_iif_socket;
    adb_thr->interface_for_adb = interface_for_adb;
    man_thr.interface_to_controller = interface_to_controller;

    // split to processing threads
#ifdef DEBUG
        log_msg(error_log, "Splitting to threads...\n", LOG_DEBUG);
#endif

    man_ret = pthread_create(&filterd_manager, NULL, filter_manager, &man_thr);

    for (int i = 0; i < thr_count; i++) {
        filt_ret[i] = pthread_create(&filterd_report[i],
                                     NULL,
                                     filter_report,
                                     &rep_thr[i]);
    }

    pthread_create(&iri_adb, NULL, iri_iif_adb_listener, adb_thr);
    // ON TERMIATE

    // wait for threads to stop
    pthread_join(filterd_manager, NULL);
    for (int i= 0; i < thr_count; i++) {
        pthread_join(filterd_report[i], NULL);
    }


    pthread_cancel(iri_adb);
    pthread_join(iri_adb, NULL);

#ifdef DEBUG
        log_msg(error_log, "Cleaning up...\n", LOG_DEBUG);
#endif

	itc_destroy(&interface_for_adb);
	itc_destroy(&interface_to_controller);

    // deallocate all structures (make current rules persistent)
    for (int i = 0; i < thr_count; i++) {
        filter_destoy(&filters[i]);
    }

    iri_iif_destroy();
    // clear filter structures
    free(filterd_report);
    free(rep_thr);
    free(filt_ret);
    free(filters);
    free(sdm_requests);

    // finalize trap
    TRAP_DEFAULT_FINALIZATION();
    return EXIT_SUCCESS;
}
