/**
 * \file pref_filter.c
 * \brief Prefix filter for filterd.
 * \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
 * \date 2014
 */
#include <stdlib.h>

#include "filter_control.h"
#include "pref_filter.h"

/**
 * \brief Procedure for initializing the prefix filter.
 * 
 * Procedure prepares prefix mask tables and initializes the list*
 * 
 * \param pf Prefix filter to be created.
 */
void pref_init(pref_filter_t *pf)
{
    memset(pf->pref_array, 0x0, sizeof(pref_item_t) * (UINT16_MAX + 1));
    pf->last_item = -1;
    for (int i = 1; i < 33; i++) {
        pf->v4_masks[i] = (0xFFFFFFFF << (32 - i));
    }
    pf->v4_masks[0] = 0x0;

    for (int i = 1; i < 64; i++) {
        pf->v6_masks[i] = (0xFFFFFFFFFFFFFFFFULL << (63 - i));
    }
    pf->v6_masks[0] = 0x0;
}

/**
 * \brief Function for comparing prefixes for sorting purposes.
 *
 * \param p Prefix 1
 * \param pref Prefix 2
 *
 * \return 0 on same result, -1 if p < pref and 1 if p > pref.
 */
int pref_compare(pref_item_t *p, pref_item_t *pref)
{
    int result = memcmp(&(p->prefix), &(pref->prefix), sizeof(sdm_ip_t));
    if (result == 0) {
        if (p->length < pref->length) {
            return -1;
        } else if (p->length > pref->length) {
            return 1;
        }
    }
    return result;
}

static inline int ipv4_pm(uint32_t ip, uint32_t pref)
{
    return (ip == pref ? 0 : (ip < pref ? -1 : 1));
}

static inline int ipv6_pm(uint64_t ip, uint64_t pref)
{
    return (ip == pref ? 0 : (ip < pref ? -1 : 1));
}

/**
 * \brief Function for matching IP address to prefix (IPv4).
 *
 * Function tries to match given IP address to a prefix in the filter using 
 * binary search.
 * 
 * \param pf Used prefix filter.
 * \param ip IP to be matched.
 * \param ip_ver Version of the input IP address
 * \param m Structure with match report.
 *
 * \return Match structure with accordingly filled fields. See filter_types.h for more information.
 */
int pref_match_v4(pref_filter_t *pf, sdm_ip_t *ip, uint8_t *m_flag, uint32_t *set_rid)
{
    if (pf->last_item == -1)
        return NO_MATCH;

    int result;
    int first = 0, mid, end;
    pref_item_t *prefixes = pf->pref_array;

    end = pf->last_item;

    while (first <= end) {
        mid = (first + end) >> 1;

        uint32_t masked;
        masked = (ip->_v4.d32[0] & pf->v4_masks[prefixes[mid].length]);
        result = ipv4_pm(masked, prefixes[mid].prefix._v4.d32[0]);

        if (result < 0) {
            end = mid - 1;
        } else if (result > 0){
            first = mid + 1;
        } else {
            *set_rid = prefixes[mid].rid;
            *m_flag |= MATCH_FOUND;
            return MATCH_FOUND;
    	}
    }
    return NO_MATCH;
}

/**
 * \brief Function for matching IP address to prefix (IPv6).
 *
 * Function tries to match given IP address to a prefix in the filter using 
 * binary search.
 * 
 * \param pf Used prefix filter.
 * \param ip IP to be matched.
 * \param ip_ver Version of the input IP address
 * \param m Structure with match report.
 *
 * \return Match structure with accordingly filled fields. See filter_types.h for more information.
 */
int pref_match_v6(pref_filter_t *pf, sdm_ip_t *ip, uint8_t *m_flag, uint32_t *set_rid)
{
    if (pf->last_item == -1)
        return NO_MATCH;

    int result;
    int first = 0, mid, end;
    pref_item_t *prefixes = pf->pref_array;

    end = pf->last_item;

    while (first <= end) {
        mid = (first + end) >> 1;

        sdm_ip_t masked;
        memset(&masked, 0x0, sizeof(sdm_ip_t));
        if (prefixes[mid].length > 64) {
            masked.v6.d64[0] = (ip->v6.d64[0] & pf->v6_masks[prefixes[mid].length - 64]);
            masked.v6.d64[1] = ip->v6.d64[1];
            result = memcmp(&masked, &(prefixes[mid].prefix), sizeof(sdm_ip_t));
        } else {
            masked.v6.d64[1] = (ip->v6.d64[1] & pf->v6_masks[prefixes[mid].length]);
            result = ipv6_pm(masked.v6.d64[1], prefixes[mid].prefix.v6.d64[1]);
        }

        if (result < 0) {
            end = mid - 1;
        } else if (result > 0){
            first = mid + 1;
        } else {
            *set_rid = prefixes[mid].rid;
            *m_flag |= MATCH_FOUND;
            return MATCH_FOUND;
    	}
    }
    return NO_MATCH;
}


/**
 * \brief Procedure for adding prefix to the filter.
 *
 * Procedure adds the given prefix to the list. To keep binary search 
 * algorithm usable procedure uses insert sort algorithm.
 *
 * \param pf Edited prefix filter.
 * \param pref Added prefix.
 */
void pref_add(pref_filter_t *pf, pref_item_t *pref)
{
    int where = 0;
    int result;

    while ((result = pref_compare(&(pf->pref_array[where]), pref)) < 0 && where <= pf->last_item) {
        if (result == 0) {
            pf->pref_array[where].rid = pref->rid;
            return;
        }
        where += 1;
    }

    if (where <= pf->last_item) {
        for (int i = pf->last_item; i >= where; i--) {
            pf->pref_array[i+1] = pf->pref_array[i];
        }
    }
       
    pf->pref_array[where] = *pref;

    pf->last_item += 1;
}

/**
 * \brief Procedure for removing item from the prefix list.
 *
 * \param pf Edited prefix list.
 * \param pref Item to remove.
 */
void pref_rem(pref_filter_t *pf, pref_item_t *pref) 
{
    unsigned int where = 0;

    while (pref_compare(&(pf->pref_array[where]), pref) != 0) {
        where += 1;
        if (where > sizeof(uint16_t))
            return;
    }

    for (int i = where; i <= pf->last_item; i++) {
        pf->pref_array[i] = pf->pref_array[i+1];
    }

    pf->last_item -= 1;
}

/**
 * \brief Procedure for destrying the prefix filter.
 *
 * \param pf Filter to be disposed.
 */
void pref_destroy(pref_filter_t *pf)
{
    memset(pf->pref_array, 0x0, sizeof(pref_item_t) * sizeof(uint16_t));
    pf->last_item = -1;
}
