/**
 * \file ini3_reporter.c
 * \brief Module for reporting intercepted frames
 * \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
 * \date 2014
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#include <unistd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "ini3d.h"
#include "ini3_reporter.h"


#undef FILE_REWIND
#undef RDWR_DIST
#undef PAYLOAD_SIZE

/**
 * \brief Macro for rewinding the file pointer.
 *
 * This macro rewinds the file pointer to the beginning of the file. It is 
 * used when either writer or reader reach FILE_FRAME_LIMIT.
 *
 * \param file File access structure to be rewound
 */
#define FILE_REWIND(file) lseek(file.fd, 0L, SEEK_SET); file.frame_index = 0

/**
 * \brief Macro for checking safety margin of the backup file. 
 *
 * This macro makes sure that both reader and writer file pointers don't
 * have enough space in case that overlapping of stored data might occur.
 *
 * \param rep File to be checked
 */
#define RDWR_DIST(rep) file->reader.frame_index - file->writer.frame_index

/**
 * \brief Macro for computing frame INI3 payload size.
 *
 * This macro is used for computing size of whole INI3 payload.
 *
 * \param frsize Size of inetrcepted frame.
 */
#define PAYLOAD_SIZE(frsize) sizeof(ini3_head_t) + frsize

static int connect_slis(ini3_reporter_t *reporter);
static int read_frame_file(backup_file_t *file, ini3_head_t *header,
                           unsigned char *message);
static int store_to_file(backup_file_t *file, ini3_head_t *header,
                         unsigned char *message, unsigned int size);
static int send_frame_direct(ini3_reporter_t *reporter, ini3_head_t *header,
                             unsigned char *message, unsigned int size);
static int send_frame_buffer(ini3_reporter_t *reporter, ini3_head_t *header,
                             unsigned char *message, unsigned int size);
static inline int check_connection(ini3_reporter_t *reporter);
static inline int check_file_full(backup_file_t *file);


/**
 * \brief Function for reporter parameters parsing
 *
 * This funtion takes command-line parameters and takes fills fields needed
 * for successful commu nication with SLIS. If some parameters were not set
 * function will use default values.
 * 
 * \param reporter Reporter instance recieving parameters.
 * \param argc Pointer to argc value
 * \param argv Pointer to argv value
 *
 * \return REP_OK if everything was set.
 * \return REP_DEFAULT if some values were set to default.
 */
int ini3_r_get_params(ini3_reporter_t *reporter, int *argc, char **argv)
{
    optind = 1;
    int ret;
    char host_set = 0, port_set = 0, probe_set = 0;
    
    while((ret = getopt(*argc, argv, "H:P:e:")) != -1) {
        switch (ret) {
            case 'H':
                host_set = 1;
                reporter->slis_param.host = optarg;
                break;

            case 'P':
                port_set = 1;
                reporter->slis_param.port = optarg;
                break;

            case 'e':
                probe_set = 1;
                reporter->slis_param.probe_id = atoi(optarg);
                break;
            default:
                continue;
        }
    }

    ret = REP_OK;

    // set defaults for items unchanged by parameters

    if (host_set == 0) {
        ret = REP_DEFAULT;
        reporter->slis_param.host = "localhost";
    }

    if (port_set == 0) {
        ret = REP_DEFAULT;
        reporter->slis_param.port = "21103";
    }

    if (probe_set == 0) {
        ret = REP_DEFAULT;
        /* TODO: check for probe_id set by CCCI for matching
           may be ommited if script will be used to run probe */
        reporter->slis_param.probe_id = 0;
    }

    return ret;
    
}

/**
 * \brief Function for INI3 reporter initialization. 
 *
 * This function initializes instance of reporter based on given parameters.
 * Function opens backup buffer file for storing intercepted frames when 
 * connection is not available. It also tries to connect to SLIS.
 * 
 * \param reporter Instance of the INI3 reporter to be initialized.
 *
 * \return REP_OK if file is opened and reporter is on-line.
 * \return REP_CONN_BROKEN if file is open but connection is not available.
 * \return REP_FILE_ERROR if backup file cannot be opened for reading/writing.
 */
int ini3_r_init(ini3_reporter_t *reporter)
{
    struct stat filestat;
    
    if (stat(BAK_PATH, &filestat) != 0)
        reporter->backup.writer.fd = open(BAK_PATH, O_CREAT | O_WRONLY | O_TRUNC, 0644);
    else
        reporter->backup.writer.fd = open(BAK_PATH, O_WRONLY | O_TRUNC); 

    reporter->backup.reader.fd = open(BAK_PATH, O_RDONLY);

    if (reporter->backup.reader.fd == -1 || reporter->backup.writer.fd == -1) {
        return REP_FILE_ERROR;
    } else {
        reporter->backup.reader.frame_index = 0;
        reporter->backup.writer.frame_index = 0;
    }

    if (connect_slis(reporter) == REP_CONN_BROKEN)
        return REP_CONN_BROKEN;
    else
        return REP_OK;
}

/**
 * \brief Function for sending intercepted frame.
 *
 * This function takes the intercepted frame and its header and sends them 
 * back to SLIS. If connection to SLIS is not available then the frame is
 * saved to the backup file. If the backup file is full the message is dropped.
 * 
 * \param reporter Instance of INI3 reporter.
 * \param[in] header INI3 payload header of the frame.
 * \param[in] frame Intercepted frame to be sent.
 *
 * \return REP_OK operation is successfully completed.
 * \return REP_CONN_BROKEN if connection was broken and frame was saved to file.
 * \return REP_FILE_ERROR if error occured during saving to file.
 * \return REP_FILE_FULL if frame was dropped because backup file is full.
 * \return REP_SEND_ERROR if error during sending to SLIS occured.
 */
int ini3_r_send(ini3_reporter_t *reporter, ini3_head_t *header,
                unsigned char *frame)
{
    int ret;
    if (check_connection(reporter) == REP_OK) {
        printf("Sending frame to SLIS ...\n");
        if (reporter->backup.frames != 0) {
            ret = send_frame_buffer(reporter, header, frame, header->blob_size);
        } else {
            ret = send_frame_direct(reporter, header, frame, header->blob_size);
        }
    } else {
        ret = store_to_file(&(reporter->backup), header, 
                            frame, header->blob_size);
        connect_slis(reporter);
    }
    
    return ret; 
}

/**
 * \brief Function for destroying INI3 reporter instance.
 *
 * This function destroys the instance of INI3 instance. It closes all opened 
 * descriptors used by the instance.
 *
 * \param[in] reporter Instance of INI3 reporter.
 */
void ini3_r_destroy(ini3_reporter_t *reporter)
{
    close(reporter->slis_socket);

    close(reporter->backup.writer.fd);
    close(reporter->backup.reader.fd);
}

/**
 * \brief Function for connecting to SLIS.
 *
 * This function takes parameters for connection and uses them to connect to
 * SLIS. This function is called by functions ini3_r_init when starting INI3d
 * and by ini3_r_send if connection was previously broken.
 *
 * \param reporter Instance of INI3d reporter for connecting.
 *
 * \return REP_OK if connection is established. 
 *         REP_CONN_BROKEN if conection is not available. 
 */
static int connect_slis(ini3_reporter_t *reporter)
{
    struct addrinfo hints, *res, *res_p;
    memset(&hints, 0x0, sizeof(struct addrinfo));

    ini3header_t conn_header = {1, 0, reporter->slis_param.probe_id};
    
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    if (getaddrinfo(reporter->slis_param.host, reporter->slis_param.port, 
                    &hints, &res) != 0) {
        return REP_CONN_BROKEN;
    }

    // try to establish connection to the server
    for (res_p = res; res_p != NULL; res_p = res_p->ai_next) {
        if ((reporter->slis_socket = 
             socket(res_p->ai_family, res_p->ai_socktype, res->ai_protocol)) == -1) {
            continue;
        }

        if (connect(reporter->slis_socket, res_p->ai_addr,
                    res_p->ai_addrlen) == -1) {
            // connection unsuccessful
            close(reporter->slis_socket);
            continue;
        } else {
            // connection successful
            write(reporter->slis_socket, &conn_header, sizeof(ini3header_t));
            freeaddrinfo(res);
            return REP_OK;
        }
    }
    freeaddrinfo(res);
    return REP_CONN_BROKEN;
}

/**
 * \brief Function for reading the frame from backup file.
 *
 * This function reads frame stored in backup file and puts the data to correct 
 * structures for sending it to SLIS.
 * 
 * \param file Structure with backup file.
 * \param header INI3 payload header to be filled.
 * \param message Buffer for frame.
 *
 * \return REP_OK if frame is read successfully.
 *         REP_FILE_ERROR if something went wrong while reading the data.
 */
static int read_frame_file(backup_file_t *file, ini3_head_t *header,
                           unsigned char *message)
{
    ssize_t read_size = 0;
    unsigned int frame_size = 0;
    off_t start_pos = lseek(file->reader.fd, 0L, SEEK_CUR);
    
    read_size = read(file->reader.fd, (unsigned char *) header, sizeof(ini3_head_t));
    
    frame_size = header->blob_size;

    read_size += read(file->reader.fd, message, sizeof(ini3_head_t));

    if (read_size != (ssize_t) PAYLOAD_SIZE(frame_size)) {
        // error while reading stored frame -- rewind
        lseek(file->reader.fd, start_pos, SEEK_SET);
        return REP_FILE_ERROR;
    }

    file->reader.frame_index += 1;
    file->reader.frame_index %= FILE_FRAME_LIMIT;
    file->frames += 1;
    if (file->reader.frame_index == 0) {
        FILE_REWIND(file->reader);
    }
    
    return REP_OK;    
}

/**
 * \brief Function for saving intercepted frame to file.
 * 
 * This function saves intercepted frame with its INI3 header to backup file
 * for later usage. This function is usually called when connetion to SLIS is
 * not available and frames will be sent later when the connection is renewed.
 * 
 * \param file File to write.
 * \param header INI3 payload header of the intercepted frame.
 * \param message Intercepted frame.
 * \param size Size of intercepted frame.
 *
 * \return REP_OK if frame is successfully writen.
 *         REP_FILE_ERROR if something went wrong when writing to file.
 */
static int store_to_file(backup_file_t *file, ini3_head_t *header,
                         unsigned char *message, unsigned int size)
{
    ssize_t stored = 0;

    // check if file buffer is full
    if (check_file_full(file))
        // buffer is full --> drop the frame
        return REP_FILE_FULL;

    // write INI3 payload to file
    stored = write(file->writer.fd, header, sizeof(ini3_head_t));
    stored += write(file->writer.fd, message, size);
        
    if (stored != (ssize_t) PAYLOAD_SIZE(size))
        return REP_FILE_ERROR;

    file->writer.frame_index += 1;
    file->writer.frame_index %= FILE_FRAME_LIMIT;
    file->frames += 1;
    if (file->writer.frame_index == 0) {
        FILE_REWIND(file->writer);
    }

    return REP_OK;
}

/**
 * \brief Function for sending frame to SLIS.
 *
 * This function takes the intercepted frame and sends it directly to SLIS
 * without checking the buffer first. It is used when backup file is empty 
 * and connection to SLIS is fully operational.
 *
 * \param reporter Instance of INI3 reporter.
 * \param header INI3 header of intercepted frame.
 * \param message Intercepted frame.
 * \param size Size of intercepted frame.
 *
 * \return REP_OK when frame is successfully sent.
 *         REP_CONN_BROKEN if connection is not available.
 *         REP_SEND_ERROR if other error happened during send operation.
 */
static int send_frame_direct(ini3_reporter_t *reporter, ini3_head_t *header,
                             unsigned char *message, unsigned int size)
{
    ssize_t sent = 0;
    if (check_connection(reporter) == REP_CONN_BROKEN)
        return REP_CONN_BROKEN;

    sent =  write(reporter->slis_socket, header, sizeof(ini3_head_t) + size);
/*    sent += write(reporter->slis_socket, message, size); */

    if (sent != (ssize_t) (PAYLOAD_SIZE(size))) {
        if (check_connection(reporter) == REP_CONN_BROKEN) {
            return REP_CONN_BROKEN;
        } else {
            return REP_SEND_ERROR;
        }
    }
    
    return REP_OK;
}

/**
 * \brief Function for sending buffered frames.
 * 
 * Function reads predefined batch of frames and sends them to SLIS. Newly 
 * recieved frame is then stored at the end of the buffer file. If the file 
 * is emptied before reaching the batch size newly recieved frame is sent as 
 * well and all following frames will be buffered when the connection breaks 
 * again.
 *
 * \param reporter INI3 reporter instance.
 * \param header header of newly arrived INI3 payload.
 * \param message Newly arrived frame.
 * \param size Size of last INI3 payload for writing/sending.
 *
 * \return 
 *  REP_OK if everything goes fine.
 *  REP_CONN_BROKEN if connection is not available or broke during send
 *  REP_SEND_ERROR if error occured during sending of the batch of saved frames.
 */
static int send_frame_buffer(ini3_reporter_t *reporter, ini3_head_t *header,
                             unsigned char *message, unsigned int size)
{
    // buffer for read frames
    unsigned char read_buffer[ETH_JUMBO_INI3];
    
    // INI3 payload header
    ini3_head_t bak_header;

    // if connection is active
    if (check_connection(reporter) == REP_CONN_BROKEN)
        return REP_CONN_BROKEN;
    
    // send predefined batch of frames
    for (int i = 0; i < FRAME_BUFFER_SEND && reporter->backup.frames > 0; i++) {
        if (read_frame_file(&(reporter->backup), &bak_header, read_buffer)
            == REP_FILE_ERROR) {
            // frame was not read
            return REP_SEND_ERROR;
        }
  
        // send payload
        if (send_frame_direct(reporter, &bak_header, read_buffer,
                              bak_header.blob_size) != REP_OK) {
            // connection broke during sendig message, seek back one frame
            lseek(reporter->backup.reader.fd,
                  0 - (PAYLOAD_SIZE(bak_header.blob_size)),
                  SEEK_CUR);
        }
        
        // decrease frame counter
        reporter->backup.frames -= 1;

        //TODO: move frame index of reader
    }

    if (reporter->backup.frames == 0) {
        // send frame to SLIS
        if (send_frame_direct(reporter, header, message, size) == REP_OK) {
            // rewind both reader and writer
            FILE_REWIND(reporter->backup.reader);
            FILE_REWIND(reporter->backup.writer);
            return REP_OK;
        } else {
            return store_to_file(&(reporter->backup), header, message, size);
        }
    } else {
        return store_to_file(&(reporter->backup), header, message, size);
    }
}

/**
 * \brief Function for checking connection to SLIS.
 *
 * This function checks if the connection to SLIS is still available.
 *
 * \param reporter Instance of INI3 reporter.
 *
 * \return REP_OK if conection is available.
 *         REP_CONN_BROKEN if connection is not available.
 */
static inline int check_connection(ini3_reporter_t *reporter)
{
    socklen_t len;
    struct sockaddr_storage addr;
    if (getpeername(reporter->slis_socket, (struct sockaddr *) &addr, &len) == -1) {
        close(reporter->slis_socket);
        return REP_CONN_BROKEN;
    } else {
        return REP_OK;
    }
}    

/**
 * \brief Function for checking backup file capacity.
 *
 * This fucntion checks the backup file capacity and reports if data
 * can be writen or the file cannot contain any more frames and different 
 * method would have to be used for frame manipulation. File is considered 
 * full when either frame count limit is reached (see FILE_FRAME_LIMIT) or 
 * next frame would be exceed the safety margin between reader and writer.
 *
 * \param file File to check.
 *
 * \return REP_OK if file has enough space for writing.
 *         REP_FILE_FULL if file is considered full.
 */
static inline int check_file_full(backup_file_t *file)
{
    if (file->frames == FILE_FRAME_LIMIT
        || (file->frames > 0 && RDWR_DIST(file) <= FRAME_SAFE_MARGIN))
        return REP_FILE_FULL;
    else
        return REP_OK;
}
