/**
 * \file ini3_reporter.h
 * \brief Module for reporting intercepted frames -- header file
 * \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
 * \date 2014
 */

#ifndef INI3_REPORTER_H
#define INI3_REPORTER_H

/**
 * \brief Path to the backup file.
 */
#define BAK_PATH "/tmp/ini3.bak"

/**
 * \brief Maximum number of frames to back up.
 */
#define FILE_FRAME_LIMIT 100

/**
 * \brief Maximum size of intercepted frame (for input buffer purpose).
 */
#define ETH_JUMBO_INI3 10000

/**
 * \brief Batch of frames to be sent when sending from backup file.
 */
#define FRAME_BUFFER_SEND 10

/**
 * \brief Size of margin to prevent data overlapping in frames.
 */
#define FRAME_SAFE_MARGIN 200

/**
 * \brief Structure for accessing the file.
 */
typedef struct file_control {
    /*@{*/
    int fd; /**< File descriptor of the backup file. */
    unsigned int frame_index; /**< Index to next writen frame. */
    /*@}*/
} file_control_t;

/**
 * \brief Structure for operating the backup file.
 */
typedef struct backup_file {
    /*@{*/
    unsigned int frames; /**< Number of currently stored frames*/
    file_control_t reader; /**< Reader access structure*/
    file_control_t writer; /**< Writer access structure*/
    /*@}*/
} backup_file_t;

/**
 * \brief Structure for SLIS connection parameters.
 */
typedef struct conn_param {
    /*@{*/
    char *host; /**< IP address/Hostname of SLIS */
    char *port; /**< SLIS port for recieving INI3 messages */
    int probe_id; /**< ID of the HI-speed probe. MUST match the ID used in CCCId */
    /*@}*/
} conn_param_t;

/**
 * \brief Structure of INI3 reporter instance.
 */
typedef struct ini3_reporter {
    /*@{*/
    conn_param_t slis_param; /**< Connection parameters of SLIS */
    int slis_socket; /**< Socket file descriptor */
    backup_file_t backup; /**< Backup file structure */
    /*@}*/
} ini3_reporter_t;

/**
 * \brief Enumeration of INI3 reporter states/errors.
 */
enum {
    /*@{*/
    REP_OK = 0, /**< Operation completed successfully. */
    REP_DEFAULT, /**< INI3 will use default settings */
    REP_INIT_ERROR, /**< Initialization error */
    REP_CONN_BROKEN, /**< Connection to SLIS not available */
    REP_SEND_ERROR, /**< Error during send operation */
    REP_FILE_ERROR, /**< Error during file I/O operation */
    REP_FILE_FULL /**< Backup file is full */
    /*@}*/
};

/**
 * \brief Function for reporter parameters parsing
 *
 * This funtion takes command-line parameters and takes fills fields needed
 * for successful commu nication with SLIS. If some parameters were not set
 * function will use default values.
 * 
 * \param reporter Reporter instance recieving parameters.
 * \param argc Pointer to argc value
 * \param argv Pointer to argv value
 *
 * \return REP_OK if everything was set.
 * \return REP_DEFAULT if some values were set to default.
 */
int ini3_r_get_params(ini3_reporter_t *reporter, int *argc, char **argv);

/**
 * \brief Function for INI3 reporter initialization. 
 *
 * This function initializes instance of reporter based on given parameters.
 * Function opens backup buffer file for storing intercepted frames when 
 * connection is not available. It also tries to connect to SLIS.
 * 
 * \param reporter Instance of the INI3 reporter to be initialized.
 *
 * \return REP_OK if file is opened and reporter is on-line.
 * \return REP_CONN_BROKEN if file is open but connection is not available.
 * \return REP_FILE_ERROR if backup file cannot be opened for reading/writing.
 */
int ini3_r_init(ini3_reporter_t *reporter);

/**
 * \brief Function for sending intercepted frame.
 *
 * This function takes the intercepted frame and its header and sends them 
 * back to SLIS. If connection to SLIS is not available then the frame is
 * saved to the backup file. If the backup file is full the message is dropped.
 * 
 * \param reporter Instance of INI3 reporter.
 * \param[in] header INI3 payload header of the frame.
 * \param[in] frame Intercepted frame to be sent.
 *
 * \return REP_OK operation is successfully completed.
 * \return REP_CONN_BROKEN if connection was broken and frame was saved to file.
 * \return REP_FILE_ERROR if error occured during saving to file.
 * \return REP_FILE_FULL if frame was dropped because backup file is full.
 * \return REP_SEND_ERROR if error during sending to SLIS occured.
 */
int ini3_r_send(ini3_reporter_t *reporter, ini3_head_t *header,
                unsigned char *frame);

/**
 * \brief Function for destroying INI3 reporter instance.
 *
 * This function destroys the instance of INI3 instance. It closes all opened 
 * descriptors used by the instance.
 *
 * \param[in] reporter Instance of INI3 reporter.
 */
void ini3_r_destroy(ini3_reporter_t *reporter);

#endif /* INI3_REPORTER_H */
