/**
 * \file ini3d_new.c
 * \brief INI3 daemon for LI HW probe
 * \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
 * \date 2014
 */  

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <sys/types.h> 
#include <errno.h>
#include <syslog.h>

#include <libtrap/trap.h>

#include "ini3d.h"
#include "ini3_reporter.h"
#include "../libs/error_logger/error_logger.h"
/* #define SZE_FRAME_OFFSET 16 */

#define DEBUG

static int stop = 0;
TRAP_DEFAULT_SIGNAL_HANDLER(stop = 1);

trap_module_info_t module_info = {
    "INI3 daemon for LI system",
    "Module for reporting captured traffic back to SLIS.\n",
    1, // prepared payloads
    0, // INI3 message
};

static void report_send_error(logger_t *log, int e)
{
    switch (e) {
        case REP_CONN_BROKEN:
            log_msg(log,"Connection to SLIS not available. Received data will be stored in backup file.",
                    LOG_WARNING);
            break;

        case REP_FILE_ERROR:
            log_msg(log,"Error when accessing backup file.",
                    LOG_ERR);
            break;

        case REP_SEND_ERROR:
            log_msg(log,"Error when trying to send data",
                    LOG_ERR);
            break;

        case REP_FILE_FULL:
            log_msg(log,"Backup file is full. Additional data will be dropped.",
                    LOG_WARNING);
            break;

    }
}

int main(int argc, char **argv)
{
    logger_t *error_log = logger_init(LOG_FILE, "INI3D", MODULES_LOG_PATH);
    if (error_log == NULL) {
        return EXIT_FAILURE;
    }

    int retval = 0;
    int trap_ret = 0;

    ini3_head_t *captured;

    const void *payload;
    uint16_t data_size;

    /* init trap */
    TRAP_DEFAULT_INITIALIZATION(argc, argv, module_info);
    TRAP_REGISTER_DEFAULT_SIGNAL_HANDLER();

    trap_ifcctl(TRAPIFC_INPUT, 0, TRAPCTL_SETTIMEOUT, TRAP_WAIT);
    trap_ifcctl(TRAPIFC_INPUT, 0, TRAPCTL_BUFFERSWITCH, 0);

    static ini3_reporter_t slis_reporter;

    if (ini3_r_get_params(&slis_reporter, &argc, argv) == REP_DEFAULT) {
        // report usage of default values
        fprintf(stderr, "[INI3d]: Not all parameters were set. These will use ");
        fprintf(stderr, "default values.\n");
        fprintf(stderr, "Defaults:\nHost: localhost\nPort: 21103\nProbe ID: 0\n");
    }

    if ((retval = ini3_r_init(&slis_reporter)) != REP_OK) {
        if (retval == REP_FILE_ERROR) {
            // report buffer file error
            log_msg(error_log,"Backup file couldn't be created/opened. Terminating ...",
                    LOG_ERR);
        } else if (retval == REP_CONN_BROKEN){
            log_msg(error_log,"Connection to SLIS not available. Backup file will be used.",
                    LOG_WARNING);
            // continue
        }
    
    }

    while (!stop) {
        // recieve intercepted frame with INI3 header in HW_DATA part 
        trap_ret = trap_recv(0, &payload, &data_size);
        TRAP_DEFAULT_GET_DATA_ERROR_HANDLING(trap_ret, continue, break);

        captured = payload;

#ifdef DEBUG
        printf("Received INI3 Report for: %hu and %hu.", captured->rid[0], captured->rid[1]);
#endif

        retval = ini3_r_send(&slis_reporter,
                             captured,
                             ((unsigned char *) payload) + sizeof(ini3_head_t));

        if (retval) {
            // report error/state during send operation
            report_send_error(error_log, retval);
        }
    }

    logger_destroy(error_log);
    ini3_r_destroy(&slis_reporter);
    trap_finalize();

    return EXIT_SUCCESS;
}
