/*
 * File: ini3d.h -- INI3 daemon for LI HW probe
 * Author: Roman Vrana, xvrana20@stud.fit.vutbr.cz
 * Date: 2013
 */  

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>

#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <sys/stat.h>
#include <fcntl.h>

#include <libtrap/trap.h>

#include "ini3d.h"

static int stop = 0;
TRAP_DEFAULT_SIGNAL_HANDLER(stop = 1);

trap_module_info_t module_info = {
    "INI3 daemon for LI system",
    "Module for reporting captured traffic back to SLIS.\n",
    1, // prepared payloads
    0, // INI3 message
};

int main(int argc, char **argv)
{

    /* init trap */
    TRAP_DEFAULT_INITIALIZATION(argc, argv, module_info);
    TRAP_REGISTER_DEFAULT_SIGNAL_HANDLER();

    trap_ifcctl(TRAPIFC_INPUT, 0, TRAPCTL_SETTIMEOUT, TRAP_WAIT);
//    trap_ifcctl(TRAPIFC_INPUT, 0, TRAPCTL_BUFFERSWITCH, 0);

    /* getopt for ini3 */

    char *ip = "localhost";
    char *port = "21103";
    uint32_t probe_id;
    int arg_ret;
    optind = 1;

    while((arg_ret = getopt(argc, argv, "H:P:e:d")) != -1) {
        switch (arg_ret) {
            case 'H':
                ip = optarg;
                break;

            case 'P':
                port = optarg;
                break;

            case 'e':
                probe_id = atoi(optarg);
                break;
            default:
                continue;
        }
    }

    const void *payload;
    uint16_t data_size;
    int trap_ret;

    struct addrinfo hints, *res, *res_p;
    int fd;
    int net_fd;
    int filed = 0;
    int net_err;
    int connected = 0;
    int registered = 0;

    unsigned back_count = 0;

    ini3header_t header;

    memset(&hints, 0, sizeof(hints));
    memset(&header, 0x0, sizeof(ini3header_t));

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    while(!stop) {

        // try to connect to SLIS server if not already connected
        if (connected == 0) {
            if ((net_err = getaddrinfo(ip, port, &hints, &res)) != 0) {
                stop = 1;
                continue;
            }

            // try to establish connection to the server
            for (res_p = res; res_p != NULL; res_p = res_p->ai_next) {
                if ((net_fd = socket(res_p->ai_family, res_p->ai_socktype, res->ai_protocol)) == -1) {
                    continue;
                }

                if (connect(net_fd, res_p->ai_addr, res_p->ai_addrlen) == -1) {
//                    fprintf(stderr, "[INI3d]: Cannot connect to SLIS. %s\n", strerror(errno));
                    close(net_fd);
                    connected = 0;
                    continue;
                } else {
                    connected = 1;
                    fd = net_fd;
                    errno = 0x0;
                    break;
                }
            }
        }

        if (connected == 0 && filed == 0) {
            fprintf(stderr, "[INI3d]: No connection to SLIS. Reports will be stored in file \"%s\".\n", INI3BAKFILE);
            fd = open("ini3payload.bak", O_WRONLY | O_CREAT, 0644);
            filed = 1;
        }

        if (connected == 1 && registered == 0) {
            registered = 1;
            header.version = 1;
            header.probe_id = probe_id;

            write(fd, (unsigned char*) &header, sizeof(ini3header_t));
        }            

        if (filed == 1 && connected == 1) {
            int bak_fd = open(INI3BAKFILE, O_RDONLY);
            ini3payload_t recovered;
            for (back_count; back_count >= 0; back_count) {
                
                // read stored payload
                read(fd, &(recovered.header.blob_size), sizeof(uint16_t));
                read(fd, &(recovered.header.sze_int), sizeof(uint8_t));
                read(fd, &(recovered.header.reserved), sizeof(uint8_t));
                read(fd, &(recovered.header.timestamp_sec), sizeof(uint32_t));
                read(fd, &(recovered.header.timestamp_ns), sizeof(uint32_t));
                read(fd, &(recovered.header.rid[0]), sizeof(uint16_t));
                read(fd, &(recovered.header.rid[1]), sizeof(uint16_t));
                read(fd, &(recovered.data), recovered.header.blob_size);

                // send stored payload to slis
                write(fd, (unsigned char*) &(recovered.header.blob_size), sizeof(uint16_t));
                write(fd, (unsigned char*) &(recovered.header.sze_int), sizeof(uint8_t));
                write(fd, (unsigned char*) &(recovered.header.reserved), sizeof(uint8_t));
                write(fd, (unsigned char*) &(recovered.header.timestamp_sec), sizeof(uint32_t));
                write(fd, (unsigned char*) &(recovered.header.timestamp_ns), sizeof(uint32_t));
                write(fd, (unsigned char*) &(recovered.header.rid[0]), sizeof(uint16_t));
                write(fd, (unsigned char*) &(recovered.header.rid[1]), sizeof(uint16_t));
                write(fd, (unsigned char*) &(recovered.data), recovered.header.blob_size);
            }

            // file sent, continue normal sending procedure
            filed = 0;
            close(bak_fd);
        }        

        trap_ret = trap_recv(0, &payload, &data_size);
        TRAP_DEFAULT_GET_DATA_ERROR_HANDLING(trap_ret, continue, break);
        if (data_size != sizeof(ini3_head_t))
            continue;
        ini3payload_t captured = *((ini3payload_t *) payload);

        trap_ret = trap_recv(0, &payload, &data_size);
        TRAP_DEFAULT_GET_DATA_ERROR_HANDLING(trap_ret, continue, break);

        /* send payload or store into file */           
        
        write(fd, (unsigned char*) &(captured.header.blob_size), sizeof(uint16_t));
        write(fd, (unsigned char*) &(captured.header.sze_int), sizeof(uint8_t));
        write(fd, (unsigned char*) &(captured.header.reserved), sizeof(uint8_t));
        write(fd, (unsigned char*) &(captured.header.timestamp_sec), sizeof(uint32_t));
        write(fd, (unsigned char*) &(captured.header.timestamp_ns), sizeof(uint32_t));
        write(fd, (unsigned char*) &(captured.header.rid[0]), sizeof(uint16_t));
        write(fd, (unsigned char*) &(captured.header.rid[1]), sizeof(uint16_t));
        write(fd, (unsigned char*) payload, captured.header.blob_size);
    }

    TRAP_DEFAULT_FINALIZATION();
    close(fd);
}
