/**
 * \file error_logger.c
 * \brief Logging library for LI probe
 * \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
 * \version 1
 * \date 2014-09-01
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>

#include "error_logger.h"

/**
* \brief Severity level name map
*/
static const char *SEVERITY_NAME[] = 
{
    /*@{*/
    [LOG_DEBUG]   = "DEBUG",    /**< Debug */
    [LOG_INFO]    = "INFO",     /**< Informative */
    [LOG_NOTICE]  = "NOTICE",   /**< Notice */
    [LOG_WARNING] = "WARNING",  /**< Warning */
    [LOG_ERR]     = "ERROR",    /**< Error */
    [LOG_CRIT]    = "CRITICAL", /**< Critical error */
    [LOG_ALERT]   = "ALERT",    /**< Alert */
    [LOG_EMERG]   = "EMERGENCY" /**< Emergency */
    /*@}*/
};


/**
* \brief Static function for generating timestamps
* This function generates timestamp for logged event in string representation
* in format "weekday month day HH:MM:SS year".
*
* \return Timestamp converted to string
*/
static char *timestamp()
{
    time_t lt = time(NULL);                                                     
    char *t = ctime(&lt);                                                       
    t[strlen(t) - 1] = '\0';

    return t;
}


/*
 * \brief Constructor of message logger.
 * This function creates new message logger. The logger is used by module 
 * to report any events during its runtime ranging from DEBUG messages 
 * to ERRORs or more severe events. Constructor automatically creates 
 * filepath given in parameter if it does not exist or opens descriptor
 * using syslog.
 *
 * \param log_type Method of logging. It should be chosen from
 *                 LOG_STDERR -- use STDERR output (used mostly for debugging)
 *                 LOG_FILE   -- use file given in path
 *                 LOG_SLOG -- use syslog
 *
 * \param module_name Name of the module using the logger for identification.
 *                    Length of the name is limited to 39 characters.
 *
 * \param path Path to log file. This parameter is only necessary when using
 *             LOG_FILE log_type. It is ignored in other methods. If set to NULL
 *             with LOG_FILE then default logging path is used
 * 
 * \return New logger on success or NULL on failure
 */

logger_t* logger_init(int log_type, const char *module_name, const char* path)
{
    logger_t *new_logger = calloc(1, sizeof(logger_t));
    char *log_file;
    int fd;
    char module_name_low[MODULE_NAME_MAX];

    if (new_logger == NULL) 
        return NULL;

    if (module_name == NULL)
        return NULL;

    switch (log_type) {
        case LOG_STDERR:
            new_logger->handle = stderr;
            break;

        case LOG_FILE:

            if (path == NULL) {
                // Create path if does not exist
                if (mkdir(DEFAULT_LOG_PATH, 0755) != 0 && errno != EEXIST)
                    return NULL;

                log_file = calloc(strlen(DEFAULT_LOG_PATH) + strlen(module_name) + 5,
                                  sizeof(char));

                for (int i = 0; i < MODULE_NAME_MAX; i++)
                    module_name_low[i] = (char) tolower((int) module_name[i]);

                sprintf(log_file, "%s%s.log", DEFAULT_LOG_PATH, module_name_low);

            } else {
                log_file = calloc(strlen(path) + strlen(module_name) + 5,
                                  sizeof(char));
                // Create path if does not exist
                if (mkdir(path, 0755) != 0 && errno != EEXIST)
                    return NULL;

                for (int i = 0; i < MODULE_NAME_MAX; i++)
                    module_name_low[i] = (char) tolower((int) module_name[i]);

                sprintf(log_file, "%s%s.log", path, module_name_low);
                
            }

            fd = open(log_file, O_WRONLY | O_CREAT, 0644);
            free(log_file);

            if (fd == -1) {
                free(new_logger);
                return NULL;
            } else {
                new_logger->handle = fdopen(fd, "w");
            }
            break;
        
        case LOG_SYSLOG:
            openlog(module_name, LOG_CONS | LOG_NDELAY, LOG_USER);
            new_logger->handle = NULL;
            break;
    }

    strncpy(new_logger->module_name, module_name, MODULE_NAME_MAX);

    return new_logger;
}

/*
 * \brief Logging function.
 * This function creates message for logged event and writes it to
 * output specified in constructor. Format of the message is:
 *
 * [MODULE|TIMESTAMP|SEVERITY_NUM|SEVERITY_STR]: MESSAGE
 * 
 * \param log Logger used for writing message.
 * \param msg Message to log.
 * \param severity Severity of event. (same levels as in syslog)
 *
 */
void log_msg(logger_t *log, const char *msg, int severity)
{
    if (log == NULL)
        return;

    if (log->handle == 0) {
        syslog(severity, "[%s|%s|%u|%s]: %s", log->module_name,
                                              timestamp(),
                                              severity,
                                              SEVERITY_NAME[severity],
                                              msg);
    } else {
        fprintf(log->handle, "[%s|%s|%u|%s]: %s\n", log->module_name,
                                                    timestamp(),
                                                    severity,
                                                    SEVERITY_NAME[severity],
                                                    msg);
        fflush(log->handle);
    }
}

/*
 * \brief Destructor of logger structure.
 * This function deallocates given logger and closes all files needed by it.
 * 
 * \param log Logger to destroy.
 */
void logger_destroy(logger_t *log)
{
    if (log == NULL)
        return;

    if (log->handle != NULL) {
        fclose(log->handle);
    } else if (log->handle == 0) {
        closelog();
    }

    free(log);
}
