/**
 * \file error_logger.h
 * \brief Logging library for LI probe -- header file
 * \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
 * \version 1
 * \date 2014-09-01
 */

#ifndef ERROR_LOGGER_H
#define ERROR_LOGGER_H

#include <syslog.h>

#define DEFAULT_LOG_PATH "/var/log/liprobe/common/"
#define MODULES_LOG_PATH "/var/log/liprobe/modules/"
#define MODULE_NAME_MAX 39
#define LOG_MAX_LEN     1024

/**
 * \brief Logger structure
 */
typedef struct logger_s {
    /*@{*/
    FILE *handle;                          /**< Output handler */ 
    char module_name[MODULE_NAME_MAX + 1]; /**< Module name for log identification */
    /*@}*/
} logger_t;

/**
* \brief Logger working modes
*/
enum {
    /*@{*/
    LOG_STDERR = 0, /**< Standard error output (mostly for debugging) */
    LOG_FILE,       /**< File output */
    LOG_SLOG        /**< Syslog output */
    /*@}*/
};

/**
 * \brief Constructor of message logger.
 * This function creates new message logger. The logger is used by module 
 * to report any events during its runtime ranging from DEBUG messages 
 * to ERRORs or more severe events. Constructor automatically creates 
 * filepath given in parameter if it does not exist or opens descriptor
 * using syslog.
 *
 * \param log_type Method of logging. It should be chosen from
 *                 LOG_STDERR -- use STDERR output (used mostly for debugging)
 *                 LOG_FILE   -- use file given in path
 *                 LOG_SLOG   -- use syslog
 *
 * \param module_name Name of the module using the logger for identification.
 *                    Length of the name is limited to 39 characters.
 *
 * \param path Path to log file. This parameter is only necessary when using
 *             LOG_FILE log_type. It is ignored in other methods. If set to NULL
 *             with LOG_FILE then default logging path is used
 * 
 * \return New logger on success or NULL on failure
 */
logger_t* logger_init(int log_type, const char *module_name, const char* path);

/**
 * \brief Logging function.
 * This function creates message for logged event and writes it to
 * output specified in constructor.
 *
 * [MODULE|TIMESTAMP|SEVERITY_NUM|SEVERITY_STR]: MESSAGE
 * 
 * \param log Logger used for writing message.
 * \param msg Message to log.
 * \param severity Severity of event. (same levels as in syslog)
 *
 */
void log_msg(logger_t *log, const char *msg, int severity);

/**
 * \brief Destructor of logger structure.
 * This function deallocates given logger and closes all files needed by it.
 * 
 * \param log Logger to destroy.
 */
void logger_destroy(logger_t *log);

#endif /* ERROR_LOGGER_H */
