##
# \file command_control.py
# \brief Commands available to control LI probe components
# \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
# \date 2014-09-04

from __future__ import print_function

import socket # communication with probe_monitor process
import os # operations with files + processes
import sys
import signal # UNIX signal constants
import readline
import threading
import subprocess as subp
from response_handler import ResponseHandler

from struct import pack
from collections import namedtuple

# ProbeConfControl class
from config_commands import ProbeConfControl


##
#Constant with configuration file path
CONFIG_FILE='/etc/liprobe/probe_config.ini'

##
#Path to probe_monitor socket
SOCK_PATH = "/tmp/probe_monitor_sck"

##
# \brief ProbeController class for controlling probe components
#
# This class represents control layer for SW components of LI probe.
# It contains running configuration during interactive mode and provides
# commands understood by the probe.
class ProbeController(object):

    ##
    # \brief Contructor of ProbeController class
    #
    # This constructor initializes probe configuration from supplied ini file.
    # It also prepares all structures for controlling the probe.
    #
    # \param interactive Switch for interactive/non-interactive mode
    #
    # \return Instance of ProbeController
    def __init__(self, interactive, config_file=None):
        self.interactive = interactive

        # create connection to probe_monitor
        try:
            self.communicator = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        except OSError as e:
            print('Cannot create connection to probe monitor.', file=sys.stderr)
            print('Check if probe_monitor process is running.', file=sys.stderr)
            if (self.interactive):
                print('Probe control commands are not available')

        try:
            self.communicator.connect(SOCK_PATH)
            if (self.interactive):
                int_mode_msg = pack("=BIxxx", 250, 0)
                self.communicator.send(int_mode_msg)
        except:
            print('Cannot create connection to probe monitor.', file=sys.stderr)
            print('Check if probe_monitor process is running.', file=sys.stderr)
            self.communicator = None
            if (self.interactive):
                print('Probe control commands are not available')

        # load configuration from supplied file or from default path
        # fallback to default setting if necessary
        if (config_file == None):
            self.probe_cfg = ProbeConfControl(CONFIG_FILE)
        else:
            self.probe_cfg = ProbeConfControl(config_file)

        self.sync = threading.Event()
        self.sync.set()
        self.reporter = ResponseHandler(self.communicator,
                                        self.sync,
                                        self.probe_cfg.get_cfg('Paths', 'log-path'))
        self.reporter_thr = threading.Thread(target=self.reporter.response_handle)
        self.reporter_thr.start()


        # dictionary of allowed interactive commands
        self.int_commands = {'start'   : self.start,
                             'stop'    : self.stop,
                             'quit'    : self.quit,
                             'exit'    : self.quit,
                             'help'    : self.help_commands,
                             'restart' : self.restart,
                             'report'  : self.report,
                             'config'  : self.config_probe,
                             'watch'   : self.watch,
                             'logging' : self.set_logging
                            }

        # dictionary of non-interactive commands
        self.commands = {'start'   : self.start,
                         'stop'    : self.stop,
                         'help'    : self.invalid,
                         'restart' : self.restart,
                         'report'  : self.report,
                         'config'  : self.probe_cfg.show_cfg
                        }

        self.cmd_hier = {'start'   : {},
                         'stop'    : {},
                         'quit'    : {},
                         'exit'    : {},
                         'help'    : {},
                         'restart' : {},
                         'watch'   : {},
                         'logging' : {},
                         'report'  : {'all'        : [],
                                      'probe'      : [],
                                      'interfaces' : [],
                                      'help'       : []
                                     },
                         'config'  : {'set' : [s.lower() for s in self.probe_cfg.cfg.sections()],
                                      'show': [],
                                      'save': [],
                                      'load': [],
                                      'help': []
                                     }
                        }

    def conn_terminate(self):
        if (self.communicator != None):
            # send EOT message
            try:
                header = pack('=BHxxxxx', 249, 0)
                cmd_msg = header
                self.communicator.send(cmd_msg)
            except (BrokenPipeError, IOError):
                pass
            finally:
                self.reporter.terminate()
                self.communicator.close()
                self.sync.set()
        

    ##
    # \brief Method for executing command
    #
    # This method takes command from user and based on this input calls
    # corresponding method from the command dictionary.
    #
    # \param command Command to execute
    # \param args Parameters of command
    def execute(self, command, args):
        if self.interactive:            
            self.int_commands[command](args)
        else:
            self.commands[command](args)

        self.sync.wait()

    ##
    # \brief Method for getting allowed interactive commands.
    # This method returns list of commands that are allowed in interactive 
    # mode.
    #
    # \return List of allowed interactive commands.
    def get_allowed_int_commands(self):
        return self.cmd_hier

    ##
    # \brief Method for starting LI probe
    #
    # This method starts all SW components of LI probe based on current 
    # running configuration. The method will fail if any of the components 
    # are already running.
    #
    # \param par Additional parameters for command method.
    def start(self, par=None):
        if (self.communicator == None):
            print('This command is not available offline')
            self.sync.set()
            return

        self.sync.clear()
        SLIS_CCCI_port = int(self.probe_cfg.get_cfg('Communication','slis-ccci-port'))
        SLIS_INI3_port = int(self.probe_cfg.get_cfg('Communication','slis-ini3-port'))
        filter_CCCI_port = int(self.probe_cfg.get_cfg('Communication','ccci-filterd-port'))
        filter_INI3_port = int(self.probe_cfg.get_cfg('Communication','ini3-filterd-port'))


        int_ifc_str = self.probe_cfg.get_cfg('GENERAL', 'probe_ifc')

    
        if (int_ifc_str.startswith('eth')):
            int_ifc = int(int_ifc_str[3:])
        elif (int_ifc_str == 'any'):
            int_ifc = 255
        elif (int_ifc_str == 'sze'):
            int_ifc = 254

        ccci_ifc = self.probe_cfg.get_cfg('Communication', 'ccci-filterd-ifc')
        ini3_ifc = self.probe_cfg.get_cfg('Communication', 'ini3-filterd-ifc')

        ccci_trap_str = self.probe_cfg.trap_ifc_code(ccci_ifc) + ';' + str(filter_CCCI_port) + ',1'
        filterd_trap_str = (self.probe_cfg.trap_ifc_code(ccci_ifc) +
                           self.probe_cfg.trap_ifc_code(ini3_ifc) + ';' +
                           self.probe_cfg.get_cfg('Communication','ccci-host') + ',' +
                           str(filter_CCCI_port) + ';' +
                           str(filter_INI3_port) + ',1')
        ini3_trap_str = (self.probe_cfg.trap_ifc_code(ini3_ifc) + ';' +
                        self.probe_cfg.get_cfg('Communication','ini3-host') + ',' +
                        str(filter_INI3_port))

        exe_path_len = len(self.probe_cfg.get_cfg('Paths','exe-path'))
        slis_host_len = len(self.probe_cfg.get_cfg('Communication','slis-host'))
        ccci_trap_len = len(ccci_trap_str)
        filterd_trap_len = len(filterd_trap_str)
        ini3_trap_len = len(ini3_trap_str)


        data = pack('=HHIBBBBBBxx{0}s{1}s{2}s{3}s{4}s'.format(exe_path_len,
                                                              slis_host_len,
                                                              ccci_trap_len,
                                                              filterd_trap_len,
                                                              ini3_trap_len),
                    SLIS_CCCI_port, SLIS_INI3_port,
                    12345,
                    slis_host_len, ccci_trap_len,
                    filterd_trap_len, ini3_trap_len,
                    exe_path_len, int_ifc,
                    self.probe_cfg.get_cfg('Paths','exe-path').encode(),
                    self.probe_cfg.get_cfg('Communication','slis-host').encode(),
                    ccci_trap_str.encode(),
                    filterd_trap_str.encode(),
                    ini3_trap_str.encode()
                   )

        header = pack('=BHxxxxx', len('start'), len(data))

        cmd_msg = header + data

        self.communicator.send(cmd_msg)
 
    ##
    # \brief Method for stopping the LI probe
    #
    # This method stops all SW components of LI probe. It also reports 
    # component was terminated safely or if it had to be killed by OS.
    #
    # \param par Additional parameters for command method.
    def stop(self, par=None):
        if (self.communicator == None):
            print('This command is not available offline')
            self.sync.set()
            return

        self.sync.clear()
        header = pack('=BHxxxxx', len('stop'), 0)
        cmd_msg = header
        self.communicator.send(cmd_msg)
 
    ##
    # \brief Method for restarting LI probe.
    #
    # This method shuts down all components of the probe and starts them again.
    # It should be used when some of the components reports ERROR state 
    # to prevent probe from doing something undesirable.
    #
    # \param par Additional parameters for command method.
    def restart(self, par=None):
        if (self.communicator == None):
            print('This command is not available offline')
            self.sync.set()
            return

        self.stop()
        self.sync.wait()
        print('stopped')
        self.start()
    ##
    # \brief Method for printing interactive commands help.
    #
    # \param par Additional parameters for command method.
    def help_commands(self, par=None):
        print()
        print('Allowed commands:')
        print()
        print("start\t\tStart LI probe.")
        print("stop\t\tStop LI probe.")
        print("quit/exit\tExit LI probe supervisor.")
        print("help\t\tPrint this help.")
        print("restart\t\tRestart LI probe.")
        print("report\t\tReport LI probe components status.")
        print("config\t\tShow configuration of LI probe components.")
        print("watch <N>\tDisplay probe health in N second intervals (0 to shutdown).")
        print()


    ##
    # \brief Method for reporting invalid command from interface
    #
    # \param par Additional parameters for command method.
    def invalid(self, par=None):
        print('I cannot do that.')
        if (self.interactive == False):
            self.conn_terminate()
            exit(1)

    ##
    # \brief Method for exiting the script
    #
    # \param par Additional parameters for command method.
    def quit(self, par=None):
        answer = input('Do you want to quit LI probe supervisor? [Y/N] ')
        if (answer.lower() == 'y'):
            if (self.probe_cfg.is_changed() == True):
                answer = input('Probe configuration may have been changed. Do you want to save it? [Y/N]')
                if (answer.lower() == 'y'):
                    self.probe_cfg.save_cfg()
            print('Shutting down')
            self.conn_terminate()
            exit(0)
        else:
            return

    ##
    # \brief Method for checking components' status
    #
    # This mehtod checks if all components are running and the checks 
    # their respective error logs if available. It also directs user to 
    # look to the error log when status of component is not perfectly ok.
    #
    # \param par Additional parameters for command method.
    def report(self, par=None):
        if (par != []):
            what = par[0]
        else:
            what = None

        if (what == 'help' or par == []):
            print('Available reporting commands')
            print()
            if (self.communicator != None):
                print('{0: <20}{1}'.format('all', 'Print status of probe software and interfaces'))
                print()
                print('{0: <20}{1}'.format('probe', 'Print only probe software status'))
                print()
            print('{0: <20}{1}'.format('interfaces', 'Print only interfaces status'))
            print()
            return

        if (self.communicator == None and (what == 'all' or what == 'probe')):
            print('This command is not available offline')
            self.sync.set()
            return

        if (what == 'all' or what == 'probe'):
            self.sync.clear()
            header = pack('=BHxxxxx', len('report'), 0)
            cmd_msg = header
            self.communicator.send(cmd_msg)
       
        self.sync.wait()
        if (what == 'all' or what == 'interfaces'):
            print()
            ibuf_raw = subp.check_output("ibufctl -A | grep -e Interface -e Packets -e Received -e Discarded -e speed | grep -v '-' | sed 's/.*: //g'",
                                         shell=True, universal_newlines=True)
            ibuf_raw = ibuf_raw.split('\n')

            for ibuf in range(0, int(len(ibuf_raw) / 5)):
                print('{0: <20}{1}'.format('Interface:', ibuf))
                print('{0: <20}{1}'.format('Link speed:', ibuf_raw[1 + ibuf*5]))
                print('{0: <20}{1}'.format('Total packets:', ibuf_raw[2 + ibuf*5]))
                print('{0: <20}{1}'.format('Received:', ibuf_raw[3 + ibuf*5]))
                print('{0: <20}{1}'.format('Discarded:', ibuf_raw[4 + ibuf*5]))
                print()

    def watch(self, par=None):
        if (self.communicator == None):
            print('This command is not available offline')
            return

        if (par == []):
            print('No timeout specified')
        else:
            try:
                timeout = int(par[0])
                header = pack('=BHxxxxx', len('monitor'), 4)
                cmd_msg = header + pack('=I', timeout)
                self.communicator.send(cmd_msg)
                self.reporter.set_timed_watch(timeout)
            except ValueError:
                print('Timeout must be numeric value')

    def set_logging(self, par=None):
        if (par == []):
            # show current logging level and help
            print('Current logging level is {0}, {1}.'.format(self.reporter.log_min_severity,
                                                               self.reporter.log_level_verbose[self.reporter.log_min_severity]))

            print('You can change logging level with command "logging" [0-7].')
        else:
            try:
                if (self.reporter.set_min_severity(int(par[0])) == 0):
                    print('Logging level set successfully.')
                else:
                    print('Cannot set logging level. Value must be in interval from 0 to 7.')
            except ValueError:
                    print('Cannot set logging level. Value must be numeric.')
        
    ##
    # \brief Method for changing configuration of the probe from interactive mode
    # 
    # This method allows user to change running configuration of the probe when
    # in interactive mode. All changes are currently not saved after exiting the 
    # script so it now serves for testing purposes.
    #
    # \param par Additional parameters for command method.
    def config_probe(self, par=None):
        command = None
        if (len(par) == 0):
            print('Specify configuration command.')
            print('Use config help to get available configuration commands')
        else:
            command = par[0]
            command_args = par[1:]

        if (command):
            if (command == 'help'):
                self.probe_cfg.show_cfg_help()
            elif (command == 'show'):
                self.probe_cfg.show_cfg()
            elif (command == 'set'):
                if (len(par) >= 2):
                    component = par[1];
                    self.probe_cfg.set_cfg(component)
            elif (command == 'save'):
                self.probe_cfg.save_cfg()
            elif (command == 'load'):
                self.probe_cfg.load_cfg()
            else:
                print('Command %s not supported' % command)
                print('Use config help to get available configuration commands')

##
# \brief Class for CLI command completion
# This class provides command completion and history for interactive mode
# interface of the script. History is saved only during run time and is not 
# kept for further use.
class CommandCompleter(object):
    def __init__(self, probectl):
        self.allowed_commands = probectl.get_allowed_int_commands()
        self.matched = []    
        
    def command_complete(self, text, state):
        response = None
        possible = []
        if state == 0:
            fullline = readline.get_line_buffer()
            l_begin = readline.get_begidx()
            l_end = readline.get_endidx()
            in_comp = fullline[l_begin:l_end]
            sep_line = fullline.split()

            if not sep_line:
                self.matched = sorted(self.allowed_commands.keys())
            else:
                try:
                    if l_begin == 0:
                        possible = sorted(self.allowed_commands.keys())
                    elif ((len(sep_line) == 1) or (len(sep_line) == 2 and in_comp)) and sep_line[0] in self.allowed_commands.keys():
                        possible = sorted(self.allowed_commands[sep_line[0]].keys())
                    elif ((len(sep_line) == 2) or (len(sep_line) == 3 and in_comp)) and sep_line[1] in self.allowed_commands[sep_line[0]].keys():
                        possible = self.allowed_commands[sep_line[0]][sep_line[1]]
                    

                    if in_comp:
                        self.matched = [c for c in possible if c.startswith(in_comp)]
                    else:
                        self.matched = possible

                except (KeyError, IndexError) as ex:
                    self.matched = []
                        

        try:
            response = self.matched[state]
        except IndexError:
            response = None

        return response

#    def short_complete(self, text):
#        if text:
#            self.matched = [s for s in self.allowed_commands if s and s.startswith(text)]
#        else:
#            self.matched = self.allowed_commands[:]
#
#        return self.matched
