##
# \file config_commands.py
# \brief Commands for configuring probe in interactive mode
# \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
# \date 2014

from __future__ import print_function

import sys
import os
import re
if sys.version_info >= (3,0):
    import configparser
else:
    import ConfigParser as configparser

from copy import deepcopy

DEFAULT_CFG_PATH = '/etc/liprobe/defaults/probe_default.ini'
DEFAULT_CFG_NAME = '/etc/liprobe/probe_config.ini'


def create_defaults():
    def_cfg = configparser.ConfigParser()
    def_cfg.add_section('GENERAL')
    def_cfg.set('GENERAL', 'probe_ifc', 'sze')
    
    def_cfg.add_section('Paths')
    def_cfg.set('Paths', 'exe-path', '/usr/local/bin/')
    def_cfg.set('Paths', 'log-path', '/var/log/liprobe/modules/')

    def_cfg.add_section('Communication')
    def_cfg.set('Communication', 'slis-host', 'localhost')
    def_cfg.set('Communication', 'ccci-host', 'localhost')
    def_cfg.set('Communication', 'ini3-host', 'localhost')
    def_cfg.set('Communication', 'slis-ccci-port', '21105')
    def_cfg.set('Communication', 'ccci-filterd-port', '65200')
    def_cfg.set('Communication', 'ini3-filterd-port', '65300')
    def_cfg.set('Communication', 'slis-ini3-port', '21103')
    def_cfg.set('Communication', 'ccci-filterd-ifc', 'unix')
    def_cfg.set('Communication', 'ini3-filterd-ifc', 'unix')

    # write new defaults
    with open(DEFAULT_CFG_PATH, 'w') as defaults:
        print('################################################################################', file=defaults)
        print('#                                                                              #', file=defaults)
        print('#     THIS IS A CONFIGURATION FILE WITH DEFAULT VALUES FOR RUNNING LI PROBE    #', file=defaults)
        print('#     FILE IS GENERATED AUTOMATICALLY IN CASE OF ACCIDENTAL DELETION           #', file=defaults)
        print('#     CHANGING ANY VALUES IN THIS FILE IS NOT ADVISED                          #', file=defaults)
        print('#                                                                              #', file=defaults)
        print('################################################################################', file=defaults)
    
        def_cfg.write(defaults)


##
# \brief 
class ProbeConfControl(object):
    def __init__(self, cfg_path):
        self.cfg = configparser.ConfigParser()
        self.def_cfg = configparser.ConfigParser()
        self.changed = False

        # check for given file
        try:
            self.cfg.readfp(open(cfg_path, 'r'))
        except (IOError, OSError):
            self.cfg = None

        # check for defaults 
        if(len(self.def_cfg.read(DEFAULT_CFG_PATH)) == 0):
            # ask for defaults file creation
            answer = None
            while (answer == None):
                print('File with default configuration was not found', file=sys.stderr)
                if sys.version_info >= (3,0):
                    answer = input('Do you want to create new default configuration? [Y/N] ')
                else:
                    answer = raw_input('Do you want to create new default configuration? [Y/N] ')
                if (answer.lower() == 'y'):
                    cself.cfg = reate_defaults()                    
                elif(answer.lower() == 'n'):
                    print('Cannot run LI probe. Default configuration does not exist', file=sys.stderr)
                    exit(99)
                else:
                    answer = None
                    print('Please answer Y or N')

        if (self.cfg == None):
            print('Configuration file not found. Falling back to default configuration.')
            self.cfg = self.def_cfg
            self.def_cfg = None

        self.cfg_change = False
        self.trap_ifc_types = {'tcp' : ('t', 'TCP Socket'),
                               'unix' : ('u', 'UNIX Socket'),
                               'null' : ('b', 'Null interface'),
                               'mem' : ('m', 'Shared memory')
                              }

        self.configurators = {'general' : self.set_general,
                              'paths' : self.set_path,
                              'communication' : self.set_communication
                             }

    def trap_ifc_code(self, ifc):
        return self.trap_ifc_types[ifc][0]

    def get_cfg(self, section, item):
        return self.cfg.get(section, item)
        
    def show_cfg_help(self, par=None):
        print('Available configuration commands')
        print()
        print('{0: <20}{1}'.format('show', 'Shows current probe configuration.'))
        print()
        print('{0: <20}{1}'.format('set <val>', 'Changes settings of specified value'))
        print()
        print('{0: <20}{1}'.format('save <filename>', 'Saves current configuration'))
        print()
        print('{0: <20}{1}'.format('load <filename>', 'Load configuration from file'))
        print()

    def show_cfg(self, par=None):
        print('General probe settings')
        print('{0: <35}{1}'.format('Intercepted Interface:', self.cfg.get('GENERAL', 'probe_ifc')))
        print()
        print()
        print('{0: <35}{1}'.format('Path to executable files:', self.cfg.get('Paths', 'exe-path')))
        print('{0: <35}{1}'.format('Paths to log files:', self.cfg.get('Paths', 'log-path')))
        print()
        print()
        print('Communication map')
        print('{0: <35}{1}'.format('SLIS <---> Probe:', self.cfg.get('Communication', 'slis-host')))
        print('{0: <35}{1}'.format('', self.cfg.get('Communication', 'slis-ccci-port') + ' (CCCI)'))
        print('{0: <35}{1}'.format('', self.cfg.get('Communication', 'slis-ini3-port') + ' (INI3)'))
        print()
        print('{0: <35}{1}'.format('Internal CCCI <---> filterd:', self.cfg.get('Communication', 'ccci-host')))
        print('{0: <35}{1}'.format('', self.cfg.get('Communication', 'ccci-filterd-port')))
        print('{0: <35}{1}'.format('', self.trap_ifc_types[self.cfg.get('Communication', 'ccci-filterd-ifc')][1]))
        print()
        print('{0: <35}{1}'.format('Internal INI3 <---> filterd:', self.cfg.get('Communication', 'ini3-host')))
        print('{0: <35}{1}'.format('', self.cfg.get('Communication', 'ini3-filterd-port')))
        print('{0: <35}{1}'.format('', self.trap_ifc_types[self.cfg.get('Communication', 'ini3-filterd-ifc')][1]))


    def is_changed(self):
        return self.changed

    def set_cfg(self, setting):
        try:
            self.changed = True
            self.configurators[setting]()
        except KeyError:
            print('Error in setting configuration.')
            print('Parameter probably cannot be set')

    def set_general(self):
        while(True):
            new_ifc = input ('Intercepted input interface [sze or ethN or any]: ')
            if (new_ifc != ''):
                if (new_ifc != 'sze' 
                    and not new_ifc.startswith('eth') 
                    and new_ifc != 'any'):
                    print('Invalid interface specification. ', end='')
                    print('Allowed interfaces are sze or any eth interface.',
                          file=sys.stderr)
                    continue
                else:
                    self.cfg.set('GENERAL', 'probe_ifc', new_ifc)
            break


    def set_path(self):
        new_path = input('New module executables folder: ')
        if (new_path != ''):
            self.cfg.set('Paths', 'exe-path',new_path)

    def set_communication(self):
        slis_host_new = input('New SLIS hostname/ip address: ')
        ccci_host_new = input('New CCCI system hostname/ip address: ')
        ini3_host_new = input('New INI3 system hostname/ip address: ')

        ccci_int_ifc = {'valid' : False, 'value' : ''}
        ini3_int_ifc = {'valid' : False, 'value' : ''}

        if (ccci_host_new != '' and ccci_host_new != 'localhost'
            or (ccci_host_new == '' and self.cfg.get('Communication', 'ccci-host') != 'localhost')):
            print('CCCId and filterd are not present on same host PC.')
            print('Internal communication interface will be automatically set to TCP')
            ccci_int_ifc = {'valid' : True, 'value' : 'tcp'}
        else:
            while (ccci_int_ifc['valid'] == False):
                ccci_int_ifc = self.input_ifc_validator(
                                'Interface for internal communication between CCCI and filterd [tcp/unix]: ',
                                ['null', 'mem'],
                                'Specified interface is not allowed'
                               )
        
        if ((ini3_host_new != '' and ini3_host_new != 'localhost')
            or (ini3_host_new == '' and self.cfg.get('Communication', 'ini3-host') != 'localhost')):
            print('INI3d and filterd are not present on same host PC.')
            print('Internal communication interface will be automatically set to TCP')
            ini3_int_ifc = {'valid' : True, 'value' : 'tcp'}
        else:
            while (ini3_int_ifc['valid'] == False):
                ini3_int_ifc = self.input_ifc_validator(
                                'Interface for internal communication between INI3d and filterd [tcp/unix]: ',
                                ['null', 'mem'],
                                'Specified interface is not allowed'
                               )
                
        slis_ccci_port = {'valid' : False, 'value' : ''}
        ccci_filterd_port = {'valid' : False, 'value' : ''}
        ini3_filterd_port = {'valid' : False, 'value' : ''}
        slis_ini3_port = {'valid' : False, 'value' : ''}

        while(slis_ccci_port['valid'] == False):
            slis_ccci_port = self.input_range_validator(
                              'Port to SLIS for CCCI [0 - 65535] :',
                              0,
                              65535,
                              'Invlaid port. Only numbers in 0 - 65535 are allowed'
                             )

        while(ccci_filterd_port['valid'] == False):
            ccci_filterd_port = self.input_range_validator(
                              'Internal port for CCCI and filterd [0 - 65535] :',
                              0,
                              65535,
                              'Invlaid port. Only numbers in 0 - 65535 are allowed'
                             )

        while(ini3_filterd_port['valid'] == False):
            ini3_filterd_port = self.input_range_validator(
                              'Internal port for INI3D and filterd [0 - 65535] :',
                              0,
                              65535,
                              'Invlaid port. Only numbers in 0 - 65535 are allowed'
                             )

        while(slis_ini3_port['valid'] == False):
            slis_ini3_port = self.input_range_validator(
                              'Port to SLIS for INI3D [0 - 65535] :',
                              0,
                              65535,
                              'Invlaid port. Only numbers in 0 - 65535 are allowed'
                             )

        ### set all values
        if (slis_host_new != ''):
            self.cfg.set('Communication', 'slis-host', slis_host_new)

        if (ccci_host_new != ''):
            self.cfg.set('Communication', 'ccci-host', ccci_host_new)

        if (ini3_host_new != ''):
            self.cfg.set('Communication', 'ini3-host', ini3_host_new)

        if (ccci_int_ifc['valid'] == True):
            self.cfg.set('Communication', 'ccci-filterd-ifc', ccci_int_ifc['value'])

        if (ini3_int_ifc['valid'] == True):
            self.cfg.set('Communication', 'ini3-filterd-ifc', ini3_int_ifc['value'])

        if (slis_ccci_port['valid'] == True):
            self.cfg.set('Communication', 'slis-ccci-port', slis_ccci_port['value'])
        
        if (ccci_filterd_port['valid'] == True):
            self.cfg.set('Communication', 'ccci-filterd-port', ccci_filterd_port['value'])
        
        if (ini3_filterd_port['valid'] == True):
            self.cfg.set('Communication', 'ini3-filterd-port', ini3_filterd_port['value'])
        
        if (slis_ini3_port['valid'] == True):
            self.cfg.set('Communication', 'ini3-slis-port', slis_ini3_port['value'])

    def save_cfg(self):
        cfg_name = input('Configuration file name [{0}]: '.format(DEFAULT_CFG_NAME))

        if (cfg_name == ''):
            cfg_name = DEFAULT_CFG_NAME

        file_patt = '(' + cfg_name + ')(_\d)*\.(ini)'
        existing = [f for f in os.listdir(os.path.expanduser('~/')) if re.match(file_patt, f)]
        if (len(existing) > 0):
            cfg_name += ('_' + len(existing))

        try:
            with open(cfg_name, 'w') as cfg_file:
                print('################################################################################',file=cfg_file)
                print('#                                                                              #',file=cfg_file)
                print('#     THIS IS A CONFIGURATION FILE WITH VALUES FOR RUNNING LI PROBE            #',file=cfg_file)
                print('#     FILE IS GENERATED AUTOMATICALLY BY CONTROL SCRIPT                        #',file=cfg_file)
                print('#     CHANGING ANY VALUES IS RECOMMENDED BY RUNNING THE CONTROL SCRIPT IN      #',file=cfg_file)
                print('#     INTERACTIVE MODE AND USING CONFIG COMMAND                                #',file=cfg_file)
                print('#                                                                              #',file=cfg_file)
                print('################################################################################',file=cfg_file)

                self.cfg.write(cfg_file)
                
                print('Configuration successfully saved.')
        except (IOError, OSError) as ex:
            print('Cannot open destination file.', file=sys.stderr)
            print('Reason: {0}'.format(ex.strerror), file=sys.stderr)

    def load_cfg(self):
        if (self.changed == True):
            answer = input('Probe configuration may have been changed. Do you want to save it? [Y/N]')
            if (answer.lower() == 'y'):
                self.save_cfg()

        print('Current working directory:\n');
        print(os.getcwd())

        load_path = input('Path to new configuration: ')

        try:
            with open(load_path, 'r') as l:
                self.cfg.readfp(l)

        except (FileNotFoundError):
            print('File {0} does not exist.'.format(load_path), file=sys.stderr)

        except (IOError, OSError) as ex:
            print('Error reading configuration file {0}'.format(load_path), file=sys.stderr)
            print('Reason: {0}'.format(ex.strerror), file=sys.stderr)

        except (configparser.ParsingError,
                configparser.MissingSectionHeaderError) as ex:
            print('Error parsing configuration file {0}'.format(load_path), file=sys.stderr)
            print('Reason: {0}'.format(ex.strerror), file=sys.stderr)
            


    def input_range_validator(self, prompt, low, high, err_msg):
        value = input(prompt)
        if (value != ''):
            try:
                x = int(value)
                if (x > high or x < low):
                    raise ValueError
                else:
                    return {'valid' : True, 'value' : value}
            except ValueError:
                print(err_msg, file=sys.stderr)
                return {'valid' : False, 'value' : value}
        else:
            return {'valid' : None, 'value' : ''}

    def input_ifc_validator(self, prompt='', restriction=[], err_msg=''):
        allowed = {k:self.trap_ifc_types[k] for k in self.trap_ifc_types if k not in restriction}
        value = input(prompt)
        if (value != ''):
            try:
                allowed[value]
                return {'valid' : True, 'value' : value}
            except KeyError:
                print(err_msg, file=sys.stderr)
                return {'valid' : False, 'value' : ''}
        else:
            return {'valid' : None, 'value' : ''}


