#!/usr/bin/env python

##
# \file probe_control.py
# \brief Main file of probe_control script
# \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
# \version 1
# \date 2014-09-04

from __future__ import print_function

import argparse # CLI arguments parsing
import readline # Modifications for input function to create more
                # shell-like interface
import sys

# ProbeController and CommandCompleter classes
from command_control import ProbeController, CommandCompleter
from config_commands import create_defaults
##
# \brief Interactive mode running point
#
# This procedure is called when the script is ran in interactive 
# mode. It provides shell like interface for controling the probe itself 
# and changing its running configuration.
#
# \param probectl Instance of ProbeControl class.

def interactive_mode(probectl):
    print('Welcome to LI probe supervisor')
    while (True):
        command = input('Command: ').split()
        try:
            if command:
                probectl.execute(command[0], command[1:])
        except KeyError:
            print('I cannot do that.',file=sys.stderr)

##
# \brief Procedure for executing command given from terminal.
#
# This procedure executes command given to probe from the terminal when 
# executing this script. It behaves more like bash scripts for daemon
# control.
#
# \param probectl Instance of ProbeControl
# \param command Command to execute

def command_mode(probectl, command):
    try:
        probectl.execute(command[0], command[1:])
    except KeyError:
        probectl.invalid()

# Configure argument parser
parser = argparse.ArgumentParser(description='Li probe supervisor')
parser.add_argument('command', nargs='?', choices=('start', 'stop', 'help', 'restart', 'report', 'config'),
                    default='help', help='command to run')
parser.add_argument('-i', dest='interactive', action='store_true')
parser.add_argument('-c', dest='config_file', default=None, action='store')
parser.add_argument('-d', dest='create_defaults', action='store_true')

args = parser.parse_known_args()

params = args[1] # unknown args will be used as params and are handled
                 # in commands
args = args[0]   # command to execute

if (args.config_file):
    print(args.config_file)
    config_file = args.config_file
else:
    config_file = None

if (args.create_defaults):
    create_defaults()
    exit(0)
    

try:
    if (args.interactive):
        # start in interactive mode and ignore other parameters
        probectl = ProbeController(True, config_file)
        cmd_comp = CommandCompleter(probectl)
        readline.set_completer(cmd_comp.command_complete)
        readline.parse_and_bind('tab: complete')

        interactive_mode(probectl)

    else:
        # perform command from terminal
        command = [args.command] + params
        probectl = ProbeController(False, config_file)
        command_mode(probectl, command)
        probectl.conn_terminate()

except KeyboardInterrupt:
    probectl.quit()
