from __future__ import print_function

import socket # communication with probe_monitor process
import os # operations with files + processes
import sys
import signal # UNIX signal constants
import select
import readline
import subprocess as subp
from struct import unpack
from collections import namedtuple

class ResponseHandler(object):
    def __init__(self, sock, sync, log_path):
        self.terminated = False
        self.timed_watch = False
        self.log_path = log_path
        self.log_min_severity = 4
        if (sock != None):
            self.sock = sock
            self.sock_buffer = 16
            self.response_check = select.poll()

            self.response_check.register(self.sock.fileno(), select.POLLIN)
            self.TIMEOUT_POLL = 2000
        else:
            self.response_check = None

        self.sync = sync
    
        self.probe_health = { 'CCCID'    : 0,
                              'Filterd' : 0,
                              'INI3D'    : 0
                            }

        # dictionary with health messages
        self.health_msg = { 0 : '\033[1;31mOFFLINE\033[0m',
                            1 : '\033[1;32mOK\033[0m',
                            3 : '\033[1;31mKILLED\033[0m',
                          }
        self.warn_msg = {7 : '\033[1;32mOK\033[0m',
                         6 : '\033[1;32mOK\033[0m',
                         5 : '\033[1;33mNOTICE\033[0m',
                         4 : '\033[1;33mWARNING\033[0m',
                         3 : '\033[1;31mERROR\033[0m',
                         2 : '\033[1;31mCRITICAL ERROR\033[0m',
                         1 : '\033[1;31mALERT\033[0m',
                         0 : '\033[1;31mEMERGENCY\033[0m'
                        }

        self.log_level_verbose = {7 : 'DEBUGGING',
                                  6 : 'INFORMATIONAL',
                                  5 : 'NOTICE',
                                  4 : 'WARNING',
                                  3 : 'ERROR',
                                  2 : 'CRITICAL ERROR',
                                  1 : 'ALERT',
                                  0 : 'EMERGENCY'
                                 }

        self.err_msg = {1 : 'Unknown command or corrupted header',
                        2 : ''
                       }

        self.err_response_format         = '=BIxxxI'
        self.state_response_format       = '=BIBxxBBB'
        self.extended_st_response_format = '=BIBxxBBBBBB'

        self.ErrResponse = namedtuple('ErrResponse', ['msg_type', 'probe_id', 'code'])
        self.StateResponse = namedtuple('StateResponse',
                                        ['type', 'probe_id', 'timed',
                                        'ccci_state', 'filter_state', 'ini3_state'])
        self.ExtStateResponse = namedtuple('ExtStateResponse',
                                           ['type', 'probe_id', 'timed',
                                            'ccci_state', 'filter_state', 'ini3_state',
                                            'ccci_log', 'filterd_log', 'ini3_log'])

        self.response_size = { 'state'     : 11,
                               'error'     : 12,
                               'ext_state' : 14
                             }

        self.response_switch = { self.response_size['state']     : self.report_state,
                                 self.response_size['error']     : self.report_error,
                                 self.response_size['ext_state'] : self.report_state
                               }
        return 

    def set_min_severity(self, val):
        if (val <= 7 and val >= 0):
            self.log_min_severity = val
            return 0
        else:
            return 1

    def set_timed_watch(self, timeout):
        if (timeout > 0):
            self.timed_watch = True
        else:
            self.timed_watch = False

    def response_handle(self):
        while (self.terminated == False and self.response_check != None):
            try:
                available = self.response_check.poll(self.TIMEOUT_POLL)
                if (len(available) != 0):
                    received = self.sock.recv(self.sock_buffer)
                    self.response_switch[len(received)](received)
                    self.sync.set()
                else:
                    continue
            except KeyError:
                pass
            except OSError:
                if (self.terminated == True):
                    pass
        return
            

    def report_error(self, data):
        print()
        response = self.ErrResponse(*unpack(self.err_response_format, data))
        print('Probe monitor reported communication error', file=sys.stderr)
        print(self.err_msg[response.code], file=sys.stderr)
        readline.redisplay()

    def report_state(self, data):
        response = None
        resp_type = None
        if (len(data) == self.response_size['ext_state']):
            response = self.ExtStateResponse(*unpack(self.extended_st_response_format, data))
        elif (len(data) == self.response_size['state']):
            response = self.StateResponse(*unpack(self.state_response_format, data))

        if (self.timed_watch == True or response.timed == 0):
            print('{0: <20}{1}'.format('CCCID', self.check_health('CCCID', response.ccci_state, response.ccci_log)))
            if (self.probe_health['CCCID'] > response.ccci_state and response.ccci_state != 0):
                self.probe_health['CCCID'] = response.ccci_state

            print('{0: <20}{1}'.format('Filterd', self.check_health('Filterd', response.filter_state, response.filterd_log)))
            if (self.probe_health['Filterd'] > response.filter_state and response.filter_state != 0):
                self.probe_health['Filterd'] = response.filter_state

            print('{0: <20}{1}'.format('INI3D', self.check_health('INI3D', response.ini3_state, response.ini3_log)))
            if (self.probe_health['INI3D'] > response.ini3_state and response.ini3_state != 0):
                self.probe_health['INI3D'] = response.ini3_state
        elif (self.timed_watch == False):
            if (self.probe_health['CCCID'] != response.ccci_state):
                print('{0: <20}{1}'.format('CCCID', self.check_health('CCCID', response.ccci_state)))
                if (self.probe_health['CCCID'] > response.ccci_state and response.ccci_state != 0):
                    self.probe_health['CCCID'] = response.ccci_state

            if (self.probe_health['Filterd'] != response.filter_state):
                print('{0: <20}{1}'.format('Filterd', self.check_health('Filterd', response.filter_state)))
                if (self.probe_health['Filterd'] > response.filter_state and response.filter_state != 0):
                    self.probe_health['Filterd'] = response.filter_state

            if (self.probe_health['INI3D'] != response.ini3_state):
                print('{0: <20}{1}'.format('INI3D', self.check_health('INI3D', response.ini3_state)))
                if (self.probe_health['INI3D'] > response.ini3_state and response.ini3_state != 0):
                    self.probe_health['INI3D'] = response.ini3_state

    def check_health(self, name, state, ext_state=0):
        last_log = ''
        logcontent = ''
        loginfo = ''
        try:
            with open(self.log_path + name.lower() + '.log', 'r') as l:
                logs = l.readlines()

                if len(logs) != 0:
                    last_log = [x.rstrip() for x in l.readlines()][-1]
                    
            if (state == 2 and last_log != ''): 
                loginfo,logcontent = last_log.rstrip().split(': ')
                loginfo = loginfo.split('|')
                if (loginfo != [] and logcontent != ''):
                    if (int(loginfo[2]) <= self.log_min_severity):
                        return self.warn_msg[int(loginfo[2])] + '\t' + logcontent
            
                return self.health_msg[1]
            else:
                return self.health_msg[state]

        except IOError as e:
            return self.health_msg[state]

    def terminate(self):
        self.terminated = True
