/**
 * \file probe_monitor.c
 * \brief LI probe control and monitoring daemon
 * \author Roman Vrana, xvrana20@stud.fit.vutbr.cz
 * \date 2015-02-16
 */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <signal.h>
#include <poll.h>
#include <fcntl.h>
#include <errno.h>
#include <inttypes.h>

#include <sys/time.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/stat.h>

#include "../../libs/error_logger/error_logger.h"
#include "probe_monitor.h"
#include "msg_parser.h"
#include "commands.h"
#include "log_reader.h"

static msg_t *todo = NULL;
static pid_t pids[3] = {0,0,0};

static pthread_mutex_t in_exec = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t exec_wait = PTHREAD_COND_INITIALIZER;

static int courier = COURIER_DEFAULT;
static int mode = 0;
static uint32_t probe_id;
static log_reader_t log_rd;

enum {
    ERR_SOCK   = -1,
    ERR_BIND   = -2,
    ERR_LISTEN = -3,
    ERR_ACCEPT = -4,
};

static int stop = 0;
static int disconnect = 0;
logger_t *evt_log;

/* Function prototypes */
int open_channel(struct pollfd *listener);
void connect_control(unsigned char *buffer, struct pollfd *listener,
                     pthread_t *execute);
void* communication(void *args);
void* execution(void *args);

void termination_handle (int s, siginfo_t *info, void *ctx)
{
    stop = 1;
    disconnect = 1;
    pthread_mutex_lock(&in_exec);
    pthread_cond_signal(&exec_wait);
    pthread_mutex_unlock(&in_exec);
    remove(PROBE_MON_PID_FILE);
}

/* create socket */

int open_channel(struct pollfd *listener)
{
    int len;
    struct sockaddr_un local;

    if ((listener->fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
        return ERR_SOCK;
    }

    local.sun_family = AF_UNIX;
    strcpy(local.sun_path, SOCK_PATH);
    unlink(local.sun_path);

    len = strlen(local.sun_path) + sizeof(local.sun_family);

    if (bind(listener->fd, (struct sockaddr *) &local, len) == -1) {
        return ERR_BIND;
    }

    if (listen(listener->fd, 5) == -1) {
        return ERR_LISTEN;
    }

    return 0;
}

/* accept / keep alive if interactive */
void connect_control(unsigned char *buffer, struct pollfd *listener,
                     pthread_t *execute)
{
    socklen_t len;
    int sel = 0, recv_len = 0;
    struct sockaddr_un remote;

    len = sizeof(remote);

    sel = poll(listener, 1, 2000);

    if (sel > 0) {
        courier = accept(listener->fd, (struct sockaddr *) &remote, &len);
        if (courier == -1) {
            courier = ERR_ACCEPT;
            return;
        }
    } else if (sel == 0) {
        courier = COURIER_DEFAULT;
        return;
    } else {
        return;
    }

    // check first message
    if ((recv_len = recv(courier, buffer, BUFF_SIZE, 0)) == -1) {
        perror("");
    }

    if (recv_len == 8 && buffer[0] == INT_MODE) {
        memset(buffer, 0x0, 8);
        mode = INT_MODE;
    }

    // create execution thread after successful
    pthread_create(execute, NULL, execution, NULL);

    pthread_detach(*execute);
}

void* communication(void *args)
{
    int len, parse_result;
    unsigned char buffer[BUFF_SIZE];
    struct pollfd listener;
    pthread_t executor;

    if (open_channel(&listener) < 0) {
        log_msg(evt_log, "Cannot open listening socket", LOG_ERR);
        stop = 1;
    }

    listener.events = POLLIN;
    
    while(!stop) {
        memset(buffer, 0x0, BUFF_SIZE);
        if (mode == 0) {
            connect_control(buffer, &listener, &executor);
        }

        if (courier < 0) {
            if (courier != COURIER_DEFAULT) {
                log_msg(evt_log, "Communication socket not opened.", LOG_ERR);
                stop = 1;
            }
            continue;
        }


        if (mode == INT_MODE) {
            if ((len = recv(courier, buffer, BUFF_SIZE, 0)) == -1) {
                log_msg(evt_log,
                        "Error in communication with front-end interface",
                        LOG_ERR);
                return NULL;
            }
        }

        msg_t *command = calloc(1, sizeof(msg_t));

        /* parse received message */
        // parse header
        if (parse_header(buffer, command) != PARSE_OK) {
            memset(buffer, 0x0, BUFF_SIZE);
            create_response(buffer, ERROR_RES, probe_id, UNKNOWN_HEADER, NULL);
            send(courier, buffer, ERR_SIZE, MSG_NOSIGNAL);

            if (mode != INT_MODE) {
                close(courier);
                courier = COURIER_DEFAULT;
            }

            continue;
        } else {
            if (command->header.type == TRANS_END) {
                mode = 0;
                disconnect = 1;
                pthread_mutex_lock(&in_exec);
                pthread_cond_signal(&exec_wait);
                pthread_mutex_unlock(&in_exec);
                close(courier);
                courier = COURIER_DEFAULT;
                disconnect = 0;
                pthread_cancel(executor);
                continue;
            }
        }
        
        // parse message
        parse_result = parse_message(command, &(buffer[8]));
        if (parse_result == PARSE_OK) {
            if (command->header.type == START_MSG)

            /* pass message to execution thread */
            pthread_mutex_lock(&in_exec);
            if (todo != NULL) {
                free(todo);
            }
            todo = command;                            
            pthread_cond_signal(&exec_wait);
            pthread_mutex_unlock(&in_exec);
            
            /* wait until response is given */
            pthread_mutex_lock(&in_exec);
            while (todo != NULL)
                pthread_cond_wait(&exec_wait, &in_exec);
            pthread_mutex_unlock(&in_exec);
        } else {
            memset(buffer, 0x0, BUFF_SIZE);
            create_response(buffer, ERROR_RES, probe_id, UNKNOWN_HEADER, NULL);
            send(courier, buffer, ERR_SIZE, MSG_NOSIGNAL);
            close(courier);
            continue;
            
        }
        if (mode != INT_MODE) {
            mode = 0;
            close(courier);
            courier = COURIER_DEFAULT;
            disconnect = 0;
        }
    }

    close(listener.fd);
    unlink(SOCK_PATH);

    return NULL;
}

void* execution(void *args)
{
    struct timespec resp_period;
    struct timeval now;

    int wait_ret = 0;
    int timed = 0;
    uint32_t mon_time = 0;
    ssize_t response_size = 0;

    unsigned char response[BUFF_SIZE];
    unsigned char states[MOD_COUNT];

    cmd_t command_vector[COMMAND_COUNT];

    populate_commands(command_vector);

    while (!stop && !disconnect) {
        memset(response, 0x0, BUFF_SIZE);
        if (mode == INT_MODE && mon_time > 0) {
            wait_ret = 0;
            // take command from recevier thread 
            gettimeofday(&now, NULL);
        
            resp_period.tv_sec = now.tv_sec + mon_time;

            pthread_mutex_lock(&in_exec);
            wait_ret = pthread_cond_timedwait(&exec_wait, &in_exec,
                                              &resp_period);
            pthread_mutex_unlock(&in_exec);
        } else {
            pthread_mutex_lock(&in_exec);
            while (todo == NULL && !disconnect)
                pthread_cond_wait(&exec_wait, &in_exec);
            pthread_mutex_unlock(&in_exec);
        }

        if (disconnect)
            return NULL;

        if (courier == COURIER_DEFAULT) {
            continue;
        }

        if (wait_ret == 0 && todo != NULL) {
            timed = 0;
            // perform given action
            switch(todo->header.type) {
                case STOP_MSG:
                    command_vector[todo->header.type](pids, states, NULL);
                    create_response(response, STATE_RES, probe_id, timed, states);
                    response_size = STATE_SIZE;
                break;

                case REPORT_MSG:
                    command_vector[todo->header.type](pids, states, &log_rd);
                    create_response(response, STATE_RES, probe_id, timed, states);
                    response_size = STATE_SIZE;
                break;

                case START_MSG:
                    command_vector[todo->header.type](todo->data, pids, NULL);
                    sleep(2);
                    command_vector[REPORT_MSG](pids, states, &log_rd);
                    create_response(response, STATE_RES, probe_id, timed, states);
                    response_size = STATE_SIZE;
                    clear_start_msg((start_msg_t *) todo->data);
                break;

                case MONITOR_MSG:
                    mon_time = *((uint32_t *) (todo->data));
                    create_response(response, SET_RES, probe_id, 0, NULL);
                    response_size = SET_SIZE;
                break;

                default:
                    create_response(response, ERROR_RES, probe_id, UNKNOWN_HEADER, NULL);
                    response_size = ERR_SIZE;
                break;
            }

            // free command structure 
            free(todo);
            todo = NULL;
        } else if (wait_ret != 0) {
            // check health periodically
            timed = 1;
            command_vector[REPORT_MSG](pids, states, &log_rd);
            create_response(response, STATE_RES, probe_id, timed, states);
            response_size = STATE_SIZE;
        } else {
            pthread_mutex_lock(&in_exec);
            pthread_cond_signal(&exec_wait);
            pthread_mutex_unlock(&in_exec);
            if (mode != INT_MODE) 
                return NULL;
            else
                continue;
        }

        /* send response message */
        send(courier, response, response_size, MSG_NOSIGNAL);
        wait_ret = 0;

        pthread_mutex_lock(&in_exec);
        pthread_cond_signal(&exec_wait);
        pthread_mutex_unlock(&in_exec);

        if (mode != INT_MODE) 
            return NULL;
    }
    return NULL;
}

int main()
{
    pid_t p = 0;
    FILE *pid_file, *probe_id_file;
    char pid_record[11];

    memset(pid_record, 0x0, 11);

#ifndef DEBUG
    evt_log = logger_init(LOG_FILE, "PROBE_MON", NULL);
#else
    evt_log = logger_init(LOG_STDERR, "PROBE_MON", NULL);
#endif

    if (evt_log == NULL) {
        perror("");
        return EXIT_FAILURE;
    }

    // check PID file existence
    if (access(PROBE_MON_PID_FILE, F_OK) != -1) {
        log_msg(evt_log,
                "Probe monitor is probably already running. "
                "Only one instance is allowed to run at a time.",
                LOG_ERR);
        return EXIT_FAILURE;
    } else {
        pid_file = fopen(PROBE_MON_PID_FILE, "w");
        if (pid_file == NULL) {
            log_msg(evt_log,
                    "Cannot start Probe Monitor -- PID file cannot be created",
                    LOG_ERR);
            return EXIT_FAILURE;
        }
    }

    // load probe_id from configuration file
    if ((probe_id_file = fopen(PROBE_ID_FILE, "r")) != NULL) {
        fscanf(probe_id_file, "%"PRIu32, &probe_id);
    } else {
        log_msg(evt_log,
                "Cannot determine probe ID. File is either corrupted or missing.",
                LOG_ERR);
        return EXIT_FAILURE;
    }    

    reader_init(&log_rd, NULL, LOG_WARNING);

#ifndef DEBUG
    p = fork();

    if (p > 0) {
        sprintf(pid_record, "%d", p);
#else
        sprintf(pid_record, "%d", getpid());
#endif
        pid_record[strlen(pid_record)] = '\n';
        fwrite(pid_record, sizeof(char), 11, pid_file);
        fclose(pid_file);

#ifndef DEBUG
        return EXIT_SUCCESS;
    } else if (p == 0) {
        fclose(pid_file);
        sleep(1);

        setsid();
        chdir("/");

        for (int i = getdtablesize(); i > 2; i--)
            close(i);

        int null = open("/dev/null", O_WRONLY);

        dup2(STDIN_FILENO, null);
        dup2(STDOUT_FILENO, null);
        dup2(STDERR_FILENO, null);
#endif
        // create threads
        pthread_t communicator;
        struct sigaction act;

        act.sa_flags = 0;
        sigemptyset(&(act.sa_mask));
        act.sa_sigaction = termination_handle;

        sigaction(SIGTERM, &act, NULL);
        sigaction(SIGINT, &act, NULL);

        act.sa_flags = 0;
        sigemptyset(&(act.sa_mask));
        act.sa_handler = SIG_IGN;
        sigaction(SIGPIPE, &act, NULL);

        pthread_create(&communicator, NULL, communication, NULL);

        pthread_join(communicator, NULL);

        return EXIT_SUCCESS;
#ifndef DEBUG
    } else {
        // error
        return EXIT_FAILURE;
    }
#endif
}

